from .common import *
from .shared import auto_get_props, add_obj


def Joker(key, module, cls):
    for n, v in getattr(cfg, key).items():
        if not v.get('IBOX_ENABLED', True):
            continue
        resname = f'{key}{n}'

        mod = __import__(f'troposphere.{module}')
        my_module = getattr(mod, module)
        my_class = getattr(my_module, cls)

        res = my_class(resname)
        auto_get_props(res, indexname=n)

        add_obj(res)



import troposphere.rds as rds

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .route53 import R53_RecordSetRDS


class RDSDBInstance(rds.DBInstance):
    def validate(self):
        if 'DBSnapshotIdentifier' not in self.properties:
            if 'Engine' not in self.properties:
                raise ValueError(
                    'Resource Engine is required in type %s'
                    % self.resource_type)

#       if 'SourceDBInstanceIdentifier' in self.properties:
#
#           invalid_replica_properties = (
#               'BackupRetentionPeriod', 'DBName', 'MasterUsername',
#               'MasterUserPassword', 'PreferredBackupWindow', 'MultiAZ',
#               'DBSnapshotIdentifier',
#           )
#
#           invalid_properties = [s for s in self.properties.keys() if
#                                 s in invalid_replica_properties]
#
#           if invalid_properties:
#                raise ValueError(
#                   ('{0} properties can\'t be provided when '
#                    'SourceDBInstanceIdentifier is present '
#                    'AWS::RDS::DBInstance.'
#                    ).format(', '.join(sorted(invalid_properties))))

        if ('DBSnapshotIdentifier' not in self.properties and
            'SourceDBInstanceIdentifier' not in self.properties) and \
            ('MasterUsername' not in self.properties or
                'MasterUserPassword' not in self.properties) and \
                ('DBClusterIdentifier' not in self.properties):
            raise ValueError(
                'Either (MasterUsername and MasterUserPassword) or'
                ' DBSnapshotIdentifier are required in type '
                'AWS::RDS::DBInstance.'
            )

        if 'KmsKeyId' in self.properties and \
           'StorageEncrypted' not in self.properties:
            raise ValueError(
                'If KmsKeyId is provided, StorageEncrypted is required '
                'AWS::RDS::DBInstance.'
            )

        nonetype = type(None)
        avail_zone = self.properties.get('AvailabilityZone', None)
        multi_az = self.properties.get('MultiAZ', None)
        if not (isinstance(avail_zone, (AWSHelperFn, nonetype)) and
                isinstance(multi_az, (AWSHelperFn, nonetype))):
            if avail_zone and multi_az in [True, 1, '1', 'true', 'True']:
                raise ValueError("AvailabiltyZone cannot be set on "
                                 "DBInstance if MultiAZ is set to true.")

        storage_type = self.properties.get('StorageType', None)
        if storage_type and storage_type == 'io1' and \
                'Iops' not in self.properties:
            raise ValueError("Must specify Iops if using StorageType io1")

        allocated_storage = self.properties.get('AllocatedStorage')
        iops = self.properties.get('Iops', None)
        if iops and not isinstance(iops, AWSHelperFn):
            if not isinstance(allocated_storage, AWSHelperFn) and \
                    allocated_storage < 100:
                raise ValueError("AllocatedStorage must be at least 100 when "
                                 "Iops is set.")
            if not isinstance(allocated_storage, AWSHelperFn) and not \
                    isinstance(iops, AWSHelperFn) and \
                    float(iops) / float(allocated_storage) > 10.0:
                raise ValueError("AllocatedStorage must be no less than "
                                 "1/10th the provisioned Iops")

        return True


class RDSDBParameterGroup(rds.DBParameterGroup):
    def __init__(self, title, dbinstance, **kwargs):
        super().__init__(title, **kwargs)
        self.Description = Sub('MYSQL %s - ${AWS::StackName}' % self.title)
        self.Family = Sub('${Engine}${EngineVersion}', **{
            'Engine': get_endvalue(f'{dbinstance}Engine'),
            'EngineVersion': Join('.', [
                Select(0,
                       Split('.', get_endvalue(f'{dbinstance}EngineVersion'))),
                Select(1,
                       Split('.', get_endvalue(f'{dbinstance}EngineVersion'))),
            ])
        })


class RDSDBSubnetGroupPrivate(rds.DBSubnetGroup):
    def setup(self):
        self.DBSubnetGroupDescription = Sub('${EnvShort}-Private')
        self.SubnetIds = Split(',', get_expvalue('SubnetsPrivate'))


class RDSDBSubnetGroupPublic(rds.DBSubnetGroup):
    def setup(self):
        self.DBSubnetGroupDescription = Sub('${EnvShort}-Public')
        self.SubnetIds = Split(',', get_expvalue('SubnetsPublic'))


def RDS_DB(key):
    for n, v in getattr(cfg, key).items():
        mapname = f'{key}{n}'

        if not v['IBOX_ENABLED']:
            continue

        try:
            resname = v['IBOX_RESNAME']
        except Exception:
            resname = mapname

        # resources
        r_DB = RDSDBInstance(mapname)
        auto_get_props(r_DB, mapname=mapname)
        # trick to keep current in use resname obj.title - need to be
        # executed after invoking auto_get_props
        r_DB.title = resname
        # trick - when providing DBSnapshotIdentifier
        # or SourceDBInstanceIdentifier some props must not exists
        for m in ['DBName',
                  'MasterUsername',
                  'MasterUserPassword']:
            setattr(r_DB, m, If(
                f'{mapname}DBInstanceSkipProperties',
                Ref('AWS::NoValue'),
                getattr(r_DB, m)))
        for m in ['SourceDBInstanceIdentifier', 'DBSnapshotIdentifier']:
            setattr(r_DB, m, If(
                f'{mapname}{m}',
                getattr(r_DB, m),
                Ref('AWS::NoValue')))

        R53_RecordSetRDS(resname)
        # trick fixed name to avoid reboot for now
        # best way should be to use a name like DBParameterGroup{Engine}
        r_PG = RDSDBParameterGroup('DBParameterGroup1', mapname)
        r_PG.Parameters = cfg.DBParameterGroup1

        add_obj([
            r_DB,
            r_PG])


def RDS_SubnetGroups(key):
    # Resources
    R_Private = RDSDBSubnetGroupPrivate('DBSubnetGroupPrivate')
    R_Private.setup()

    R_Public = RDSDBSubnetGroupPublic('DBSubnetGroupPublic')
    R_Public.setup()

    # Outputs
    O_Private = Output('DBSubnetGroupPrivate')
    O_Private.Value = Ref('DBSubnetGroupPrivate')
    O_Private.Export = Export('DBSubnetGroupPrivate')

    O_Public = Output('DBSubnetGroupPublic')
    O_Public.Value = Ref('DBSubnetGroupPublic')
    O_Public.Export = Export('DBSubnetGroupPublic')

    add_obj([
        R_Private,
        R_Public,
        O_Private,
        O_Public])



import python_minifier
import troposphere.ssm as ssm
import troposphere.policies as pol

from .common import *


# S - PARAMETERS #
class Parameter(Parameter):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Type = 'String'
        self.Default = ''


class SSMParameter(ssm.Parameter):
    Type = 'String'
# E - PARAMETERS #


def stack_add_res():
    for n, v in cfg.Parameters.items():
        # Automatically create override conditions for parameters
        if not n.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION):

            if n.endswith('InstanceType'):
                default = 'default'
            elif n == 'SecurityGroups':
                default = cfg.SECURITY_GROUPS_DEFAULT
            else:
                default = ''

            condition = {
                f'{n}Override': Not(Equals(Select(0, Ref(n)), default))
            } if v.Type == 'CommaDelimitedList' else {
                f'{n}Override': Not(Equals(Ref(n), default))
            }

            add_obj(condition)
        # End
        cfg.template.add_parameter(v)

    cfg.Parameters.clear()

    for n, v in cfg.Conditions.items():
        cfg.template.add_condition(n, v)
    cfg.Conditions.clear()

    for n, v in cfg.Mappings.items():
        cfg.template.add_mapping(n, v)
    cfg.Mappings.clear()

    for n, v in cfg.Resources.items():
        cfg.template.add_resource(v)
    cfg.Resources.clear()

    for n, v in cfg.Outputs.items():
        add_objoutput(v)
        cfg.template.add_output(v)
    cfg.Outputs.clear()


def add_obj(obj):
    if isinstance(obj, list):
        for n in obj:
            add_obj(n)
    elif isinstance(obj, Parameter):
        cfg.Parameters[obj.title] = obj
    elif isinstance(obj, dict):
        cfg.Conditions.update(obj)
    elif isinstance(obj, Output):
        cfg.Outputs[obj.title] = obj
    else:
        cfg.Resources[obj.title] = obj


def add_objoutput(res):
    try:
        iboxprops = res.IBOX_PROPS
    except Exception:
        pass
    else:
        mapname = iboxprops['MAP'][res.title]

        if isinstance(res.Value, str):
            join_list = []
            for n in res.Value.split():
                n = n.strip()
                if n.startswith('${'):
                    n = n.strip('${}')
                    obj = iboxprops[f'{mapname}{n}'][0]
                    propname = iboxprops[f'{mapname}{n}'][1]
                    n = getattr(obj, propname)
                join_list.append(n)

            res.Value = Join('', join_list)
        else:
            # output use relative resource condition if do not have one
            # i do this only if Value is not a string
            # i need Condition only if using Ref for getting the resource
            try:
                res.Condition
            except Exception:
                try:
                    cond = iboxprops[f'{mapname}Condition'][0].Condition
                except Exception:
                    pass
                else:
                    res.Condition = cond

        del res.properties['IBOX_PROPS']


def do_no_override(action):
    if action is True:
        cfg.no_override = True
    else:
        cfg.no_override = False


def get_endvalue(param, ssm=False, condition=False, nocondition=False,
                 nolist=False, inlist=False, split=False, issub=False,
                 strout=False, fixedvalues=None, mapinlist=False):
    if not fixedvalues:
        # set default if not defined
        fixedvalues = cfg.fixedvalues

    def _get_overridevalue(param, value, condname=None):
        if param not in cfg.Parameters and condname in cfg.Parameters:
            param = condname
        if (cfg.no_override is False and param in cfg.Parameters and not
                param.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION)):
            override_value = If(f'{param}Override', Ref(param), value)
        else:
            override_value = value

        return override_value

    def _get_value():
        # if param in fixedvalues means its value do not changes
        # based on Env/Region so hardcode the value in json, ...
        if param in fixedvalues:
            value = fixedvalues[param]
            # check if value start with method and use eval to run code
            if isinstance(value, list):
                value = [
                    eval(r) if str(r).startswith(
                        cfg.EVAL_FUNCTIONS_IN_CFG) else r for r in value
                ]
            if isinstance(value, str):
                value = (
                    eval(value.replace('\n', ''))
                    if value.startswith(cfg.EVAL_FUNCTIONS_IN_CFG)
                    else value)
        # ... otherway use mapping
        else:
            value = FindInMap(Ref('EnvShort'), Ref('AWS::Region'), param)

        if strout is True and isinstance(value, int):
            value = str(value)

        if nolist is True and isinstance(value, list):
            value = ','.join(value)

        if issub:
            value = Sub(value)

        return value

    value = _get_value()

    if mapinlist is not False:
        value = Select(mapinlist, value)

    if condition or nocondition:
        if condition is True or nocondition is True:
            condname = param
        elif condition:
            condname = condition
        elif nocondition:
            condname = nocondition

        value = _get_overridevalue(param, value, condname)

        value = If(
            condname,
            value if condition else Ref('AWS::NoValue'),
            Ref('AWS::NoValue') if condition else value,
        )
    else:
        value = _get_overridevalue(param, value)

    if split is not False:
        value = Select(split, Split(',', value))

    if inlist is not False:
        value = Select(inlist, value)

    return value


def get_expvalue(param, stack=False, prefix=''):
    v = ''
    if stack:
        v = ImportValue(
            Sub('%s-${%s}' % (param, stack), **{stack: get_endvalue(stack)})
        )
    elif not isinstance(param, str):
        v = ImportValue(
            Sub('%s${ImportName}' % prefix, **{'ImportName': param})
        )
    else:
        v = ImportValue(param)

    return v


def get_subvalue(substring, subvar, stack=False):
    submap = {}
    found = substring.find('${')

    while found != -1:
        posindex = found + 2
        myindex = substring[posindex]
        mytype = substring[posindex + 1]
        # M = Mapped, E = Exported
        if myindex.isdigit() and mytype in ['M', 'E']:
            listitem = (
                subvar[int(myindex) - 1] if isinstance(subvar, list)
                else subvar)
            stackitem = (
                stack[int(myindex) - 1] if isinstance(stack, list)
                else stack)
            if mytype == 'M':
                submap[listitem] = get_endvalue(listitem)
            else:
                submap[listitem] = get_expvalue(listitem, stackitem)

            substring = substring.replace(
                '${%s%s}' % (myindex, mytype), '${%s}' % listitem)

        found = substring.find('${', posindex)

    v = Sub(substring, **submap)

    return v


def get_resvalue(resname, propname):
    res = cfg.Resources[resname]

    loc = propname.find('.')
    while loc > 0:
        prop = propname[0:loc]
        res = getattr(res, prop)
        propname = propname[loc + 1:]
        loc = propname.find('.')

    return getattr(res, propname)


def get_dictvalue(key):
    if isinstance(key, list):
        value = [get_dictvalue(k) for k in key]
    elif isinstance(key, dict):
        value = {i: get_dictvalue(k) for i, k in key.items()}
    else:
        value = eval(key) if key.startswith(
            cfg.EVAL_FUNCTIONS_IN_CFG) else key

    return value


def get_condition(cond_name, cond, value2, key=None, OrExtend=[],
                  mapinlist=False, nomap=None):
    # record current state
    override_state = cfg.no_override
    do_no_override(True)

    key_name = key if key else cond_name
    if isinstance(key, FindInMap):
        map_name = key.data['Fn::FindInMap'][0]
        key_name = key.data['Fn::FindInMap'][1]
        value_name = key.data['Fn::FindInMap'][2]
        if not value_name and cond_name:
            value_name = cond_name

        value1_param = FindInMap(map_name, Ref(key_name), value_name)
        value1_map = FindInMap(map_name, get_endvalue(key_name), value_name)
    elif isinstance(key, Select):
        select_index = key.data['Fn::Select'][0]
        select_list = key.data['Fn::Select'][1]

        if 'Fn::Split' in select_list.data:
            split_sep = select_list.data['Fn::Split'][0]
            key_name = select_list.data['Fn::Split'][1]
            select_value_param = Split(split_sep, Ref(key_name))
            select_value_map = Split(split_sep, get_endvalue(key_name))
        else:
            select_value_param = select_list
            select_value_map = get_endvalue(select_list)

        value1_param = Select(
            select_index, select_value_param)
        value1_map = Select(
            select_index, select_value_map)
    else:
        value1_param = Ref(key_name)
        # Used new param "mapinlist" when you have a mapped value in a list
        # but multiple values as override parameters
        if mapinlist:
            value1_map = get_endvalue(mapinlist[0], mapinlist=mapinlist[1])
        else:
            value1_map = get_endvalue(key_name)

    # if beginning state was False set it back
    if not override_state:
        do_no_override(False)

    eq_param = Equals(value1_param, value2)
    eq_map = Equals(value1_map, value2)

    if cond == 'equals':
        cond_param = eq_param
        cond_map = eq_map
    elif cond == 'not_equals':
        cond_param = Not(eq_param)
        cond_map = Not(eq_map)

    if (key_name in cfg.Parameters and
            not key_name.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION) and
            not nomap):
        key_override = f'{key_name}Override'
        condition = Or(
            And(
                Condition(key_override),
                cond_param,
            ),
            And(
                Not(Condition(key_override)),
                cond_map,
            )
        )
        if OrExtend:
            condition.data['Fn::Or'].extend(OrExtend)
    elif nomap:
        condition = cond_param
    else:
        condition = cond_map

    if cond_name:
        condition = {cond_name: condition}

    return condition


def import_lambda(name):
    TK_IN_LBD = 'IBOX_CODE_IN_LAMBDA'
    parent_dir_name = os.path.dirname(os.path.realpath(__file__))
    lambda_file = os.path.join(os.getcwd(), 'lib/lambdas/%s.code' % name)
    if not os.path.exists(lambda_file):
        lambda_file = os.path.join(parent_dir_name, 'lambdas/%s.code' % name)
    try:
        with open(lambda_file, 'r') as f:
            fdata = f.read()

            try:
                # try to minify using python_minifier
                code = python_minifier.minify(fdata)
            except Exception:
                code = fdata

            if len(code) > 4096:
                logging.warning(f'Inline lambda {lambda_file} > 4096')

            code_lines = code.splitlines(keepends=True)

            file_lines = []
            # parse lambda code for Token IBOX CODE
            for x in code_lines:
                if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    value = eval(x)
                elif x.startswith(TK_IN_LBD):
                    value = '"'
                elif TK_IN_LBD in x:
                    # parse minified code
                    tks = x.split(TK_IN_LBD)
                    file_lines.extend([
                        f'{tks[0]}', eval(tks[1]), tks[2]])
                    continue
                else:
                    value = ''.join(x)

                file_lines.append(value)

            return(file_lines)

    except IOError:
        logging.warning(f'Lambda code {name} not found')
        exit(1)


def auto_get_props(obj, mapname=None, key=None, rootdict=None, indexname=''):
    # IBOX_RESNAME can be used in yaml and resolved inside get_endvalue
    global IBOX_RESNAME, IBOX_MAPNAME, IBOX_INDEXNAME
    IBOX_RESNAME = obj.title
    IBOX_MAPNAME = mapname
    IBOX_INDEXNAME = indexname

    # create a dict where i will put all property with a flat hierarchy
    # with the name equals to the mapname and the relative value.
    # Later i will assign this dict to the relative output object using
    # IBOX_OUTPUT
    IBOX_PROPS = {'MAP': {}}

    def _iboxif(if_wrapper, mapname, value):
        condname = if_wrapper[0].replace('IBOX_MAPNAME_', mapname)
        condname = condname.replace('_', IBOX_RESNAME)
        condvalues = []
        for i in if_wrapper[1:3]:
            if isinstance(i, str) and i.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                v = eval(i)
            else:
                v = i
            condvalues.append(v)

        if condvalues[0] == 'IBOX_IFVALUE':
            value = If(condname, value, condvalues[1])
        else:
            value = If(condname, condvalues[0], value)

        return value

    def _get_obj(obj, key, obj_propname, mapname):
        props = obj.props
        mapname_obj = f'{mapname}{obj_propname}'

        def _get_obj_tags():
            prop_list = []
            for k, v in key[obj_propname].items():
                try:
                    v['Value']
                except Exception:
                    pass
                else:
                    v['Value'] = get_endvalue(f'{mapname_obj}{k}Value')

                try:
                    if_wrapper = v['IBOX_IF']
                except Exception:
                    pass
                else:
                    del v['IBOX_IF']
                    v = _iboxif(if_wrapper, mapname, v)

                prop_list.append(v)

            obj_tags = Tags()
            obj_tags.tags = prop_list

            return obj_tags

        # trick (bad) to detect attributes as CreationPolicy and UpdatePolicy
        if obj_propname in ['CreationPolicy', 'UpdatePolicy']:
            prop_class = getattr(pol, obj_propname)
        else:
            prop_class = props[obj_propname][0]

        if isinstance(prop_class, type):
            # If object already have that props, object class is
            # the existing already defined object,
            # so extend its property with auto_get_props
            if obj_propname in obj.properties:
                prop_obj = obj.properties[obj_propname]
            else:
                prop_obj = prop_class()

            if prop_class.__bases__[0].__name__ in ['AWSProperty',
                                                    'AWSAttribute']:
                _populate(prop_obj, key=key[obj_propname], mapname=mapname_obj)
            elif prop_class.__name__ == 'dict':
                prop_obj = get_dictvalue(key[obj_propname])
            elif prop_class.__name__ == 'Tags':
                prop_obj = _get_obj_tags()

            return prop_obj

        elif isinstance(prop_class, tuple) and any(n in prop_class for n in [
                Tags, asgTags]):
            prop_obj = _get_obj_tags()

            return prop_obj

        elif (isinstance(prop_class, list) and
                isinstance(prop_class[0], type) and
                prop_class[0].__bases__[0].__name__ == 'AWSProperty'):
            prop_list = []
            prop_class = props[obj_propname][0][0]
            for o, v in key[obj_propname].items():
                if o == 'IBOX_IF':
                    # element named IBOX_IF must no be parsed
                    # is needed for wrapping whole returned obj in _populate
                    continue
                name_o = str(o)
                mapname_o = f'{mapname_obj}{name_o}'
                prop_obj = prop_class()

                _populate(prop_obj, key=v, mapname=mapname_o)

                # trick to wrapper single obj in If Condition
                try:
                    if_wrapper = v['IBOX_IF']
                except Exception:
                    prop_list.append(prop_obj)
                else:
                    prop_list.append(_iboxif(if_wrapper, mapname, prop_obj))

            return prop_list

    def _populate(obj, key=None, mapname=None, rootdict=None):
        if not mapname:
            mapname = obj.title
        if rootdict:
            key = rootdict
            mapname = ''
        if not key:
            key = getattr(cfg, mapname)

        def _try_PCO_in_obj(key):
            def _parameter(k):
                for n, v in k.items():
                    n = n.replace('{IBOX_INDEXNAME}', IBOX_INDEXNAME)
                    n = n.replace('_', IBOX_RESNAME)
                    parameter = Parameter(n)
                    _populate(parameter, rootdict=v)
                    add_obj(parameter)

            def _condition(k):
                for n, v in k.items():
                    if IBOX_MAPNAME:
                        n = n.replace('IBOX_MAPNAME', IBOX_MAPNAME)
                    n = n.replace('{IBOX_INDEXNAME}', IBOX_INDEXNAME)
                    n = n.replace('_', IBOX_RESNAME)
                    condition = {n: eval(v)}
                    add_obj(condition)

            def _output(k):
                for n, v in k.items():
                    n = n.replace('{IBOX_INDEXNAME}', IBOX_INDEXNAME)
                    n = n.replace('_', IBOX_RESNAME)
                    output = Output(n)
                    _populate(output, rootdict=v)
                    # alter troposphere obj and add IBOX property
                    output.props['IBOX_PROPS'] = (dict, False)
                    # starting from troposphere 3.0 propnames is a set
                    output.propnames = list(output.propnames) + ['IBOX_PROPS']
                    # assign auto_get_props populated obj to IBOX property
                    output.IBOX_PROPS = IBOX_PROPS
                    output.IBOX_PROPS['MAP'][n] = mapname
                    add_obj(output)

            func_map = {
                'IBOX_PARAMETER': _parameter,
                'IBOX_CONDITION': _condition,
                'IBOX_OUTPUT': _output,
            }
            for pco in list(func_map.keys()):
                try:
                    k = key[pco]
                except Exception:
                    pass
                else:
                    func_map[pco](k)

        def _auto_PO(name, conf, mode, value=''):
            if mode == 'p':
                parameter_base = {'Description': 'Empty for mapped value'}
                parameter_base.update(conf)
                parameter = {'IBOX_PARAMETER': {
                    f'{mapname}{name}': parameter_base}}
                _try_PCO_in_obj(parameter)
            if mode == 'o':
                if isinstance(value, int):
                    # Output value must be a string
                    value = str(value)
                output_base = {'Value': value}
                output_base.update(conf)
                output = {'IBOX_OUTPUT': {
                    f'{mapname}{name}': output_base}}
                _try_PCO_in_obj(output)

        # Allowed object properties
        props = list(vars(obj)['propnames'])
        props.extend(vars(obj)['attributes'])

        # Parameters, Conditions, Outpus in yaml cfg
        _try_PCO_in_obj(key)

        for propname in dict(
                list(key.items()) +
                list(dict.fromkeys(props, None).items())).keys():
            # iterate over both obj props and key - using key order

            # IBOX_AUTO_PO
            for n in [f'{propname}.IBOX_AUTO_PO',
                      f'{propname}.IBOX_AUTO_P']:
                if n in key:
                    # Automatically create parameter
                    _auto_PO(propname, key[n], 'p')

            # IBOX_PCO
            ibox_pco = f'{propname}.IBOX_PCO'
            if ibox_pco in key:
                # If there is a key ending with {prop}.IBOX_PCO process it
                _try_PCO_in_obj(key[ibox_pco])

            # IBOX_CODE
            ibox_code = f'{propname}.IBOX_CODE'
            if ibox_code in key:
                # If there is a key ending with {prop}.IBOX_CODE
                # eval it and use it as prop value.
                # Usefull if a str value need to be processed by a code.
                # (like in autoscaling-scheduledactions.yml)
                value = eval(key[ibox_code])
            elif propname in key and propname in props:
                # there is match between obj prop and a dict key
                key_value = key[propname]

                # set value
                if isinstance(key_value, dict):
                    # key value is a dict, get populated object
                    value = _get_obj(obj, key, propname, mapname)
                elif (isinstance(key_value, str)
                        and key_value.startswith('IBOX_RESNAME')):
                    # replace IBOX_RESNAME string with IBOX_RESNAME value
                    value = key_value.replace('IBOX_RESNAME', IBOX_RESNAME)
                elif rootdict:
                    # rootdict is needed by lib/efs.py EFS_FileStorage SGIExtra
                    # where is passed as a dictionary to parse for parameters
                    value = get_endvalue(
                        f'{mapname}{propname}', fixedvalues=rootdict)
                elif (isinstance(key_value, str)
                        and key_value.startswith('get_endvalue(')):
                    # Usefull to migrate code in yaml using auto_get_props
                    # get_endvalue is used only when migrating code
                    value = eval(key_value)
                else:
                    value = get_endvalue(f'{mapname}{propname}')

                if key_value == 'IBOX_IFCONDVALUE':
                    # auto add condition and wrap value in AWS If Condition
                    add_obj(get_condition(f'{mapname}{propname}',
                                          'not_equals', 'IBOX_IFCONDVALUE'))
                    value = If(
                        f'{mapname}{propname}',
                        value,
                        Ref('AWS::NoValue'))

                # trick to wrapper recursed returned value in If Condition
                try:
                    if_wrapper = key_value['IBOX_IF']
                except Exception:
                    pass
                else:
                    value = _iboxif(if_wrapper, mapname, value)

                if propname == 'Condition' and not isinstance(value, str):
                    # Avoid intercepting a Template Condition
                    # as a Resource Condition
                    continue
            else:
                # NO match between propname and cfg keys
                continue

            # Finally set obj property
            try:
                setattr(obj, propname, value)
            except TypeError:
                pass
            else:
                # populate IBOX_PROPS dict
                IBOX_PROPS[f'{mapname}{propname}'] = (obj, propname)

                # Automatically create output
                for n in [f'{propname}.IBOX_AUTO_PO',
                          f'{propname}.IBOX_AUTO_O']:
                    if n in key:
                        _auto_PO(propname, key[n], 'o', value)

        # title override
        try:
            obj.title = key['IBOX_TITLE']
        except Exception:
            pass

    _populate(obj, key, mapname, rootdict)
    return obj


def auto_build_obj(obj, key, name=None):
    props = vars(obj)['propnames']
    classname = obj.__class__
    for resname, resvalue in key.items():
        final_obj = classname(resname)
        if not name:
            name = final_obj.__class__.__name__
        auto_get_props(final_obj, f'{name}{resname}')

        add_obj(final_obj)


def change_obj_data(obj, find, value):
    if isinstance(obj, list):
        for n, v in enumerate(obj):
            change_obj_data(obj[n], find, value)
    elif isinstance(obj, If):
        obj_if = obj.data['Fn::If']
        for n, v in enumerate(obj_if):
            if isinstance(v, Ref) and v.data['Ref'] == find:
                obj_if[n] = value


def gen_random_string():
    length = 16
    char_set = f'{string.ascii_letters}{string.digits}'
    if not hasattr(gen_random_string, "rng"):
        # Create a static variable
        gen_random_string.rng = random.SystemRandom()
    secret_string = ''.join(
        [gen_random_string.rng.choice(char_set) for _ in range(length)])

    return secret_string


def clf_compute_order(pattern):
    base_ord = 1

    for s, w in cfg.CLF_PATH_PATTERN_REPLACEMENT.items():
        pattern = pattern.replace(w, s)

    n_star = 0
    for n, v in enumerate(pattern):
        if v == '?':
            base_ord += 0.0001
        elif v == '*':
            if len(pattern) == n + 1 and n_star != 0:
                # v is last char but not the only star char
                base_ord += 0.0005
            elif n_star != 0:
                base_ord -= 0.000001*n
            else:
                base_ord += 1000/n
                n_star += 1
        elif n_star != 0:
            base_ord -= 0.001

    cfg.dbg_clf_compute_order[pattern] = base_ord

    return base_ord


def camel_to_snake(data):
    out = ''
    skip_next = False
    for n, v in enumerate(data):
        if skip_next:
            skip_next = False
            continue
        if v.isupper() and data[n+1].isupper():
            out += f'_{v}{data[n+1]}'
            skip_next = True
        elif v.isupper():
            out += f'_{v}' if out else v
        else:
            out += v

    return out.upper()



from . import cfg
from .shared import stack_add_res
from . import (mappings, lambdas, securitygroup, loadbalancing,
               autoscaling, iam, codedeploy, route53, events, cloudfront,
               sqs, sns, ecs, ecr, s3, waf, vpc, dynamodb, kms, rds, efs,
               elasticache, cloudformation, logs, apigateway,
               joker, __version__)


def execute_method(RP_cmm):
    for k, v in cfg.CFG_TO_FUNC.items():
        func_name = v['func']
        module_name = v['module']
        module = globals()[module_name]

        if k in list(RP_cmm.keys()):
            RP_value = RP_cmm[k]
            if isinstance(RP_value, str) and RP_value == 'IBOX_SKIP_FUNC':
                continue
            if isinstance(func_name, list):
                for n in func_name:
                    getattr(module, n)(key=k)
                continue
            stacktype_func = f'{func_name}{cfg.stacktype.upper()}'
            if stacktype_func in dir(module):
                getattr(module, stacktype_func)(key=k)
            elif func_name in dir(module):
                getattr(module, func_name)(key=k)
            elif module_name == 'joker':
                # for resources that can be built using only auto_get_props
                joker.Joker(key=k, module=func_name[0], cls=func_name[1])

    stack_add_res()


def generate():
    cfg.func_envrole = cfg.envrole.replace('-', '_')

    try:
        execute_method(cfg.RP_cmm)
    except Exception:
        # in case of error call stack_add_res to empty the cfg Parameters
        # Conditions Mappings Resources Outputs  dicts. This is needed cause
        # get_template "restore base cfg" section do not recursive remove
        # "sub-objetcs" or to say it in a better way:
        # dict clear methods reset the contents,
        # but all references to the same instance will still be present.
        stack_add_res()
        raise

    version_text = f'(Generated by awsibox {__version__})'
    cfg.template.set_description(
        f'{cfg.envrole} {version_text} [{cfg.stacktype}]')
    cfg.template.set_version('2010-09-09')

    return cfg.template



import troposphere.apigateway as agw
import troposphere.route53 as r53

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionApiGateway
from .iam import IAMPolicyApiGatewayPrivate


class ApiGatewayAccount(agw.Account):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.CloudWatchRoleArn = get_expvalue('RoleApiGatewayCloudWatch')


class ApiGatewayResource(agw.Resource):
    def __init__(self, title, key, stage, **kwargs):
        super().__init__(title, **kwargs)
        mapname = f'ApiGatewayStage{stage}ApiGatewayResource'

        auto_get_props(self)
        self.RestApiId = Ref('ApiGatewayRestApi')

        # if stage override ApiGatewayResource specs, get it
        try:
            getattr(cfg, mapname)
        except Exception:
            pass
        else:
            auto_get_props(self, mapname)


class ApiGatewayMethod(agw.Method):
    def __init__(self, title, key, basename, name, stage, **kwargs):
        super().__init__(title, **kwargs)
        mapname = (f'ApiGatewayStage{stage}'
                   f'ApiGatewayResource{basename}Method{name}')

        try:
            getattr(cfg, mapname)
        except Exception:
            pass
        else:
            auto_get_props(self, mapname)

        auto_get_props(self)
        self.RestApiId = Ref('ApiGatewayRestApi')
        self.ResourceId = Ref(f'ApiGatewayResource{basename}')

        # If Uri is a lambda self.Integration.Uri will be like:
        # 'arn:aws:apigateway:${AWS::Region}:lambda:path/2015-03-31/functions/${LambdaName.Arn}/invocations'
        # i need to append stage version (Ex. v1) to the Lambda Name so that
        # ${LambdaName.Arn} -> ${LambdaNameV1.Arn}
        if ':lambda:' in self.Integration.Uri:
            iu = self.Integration.Uri
            dot_found = iu.rfind('.')
            before_dot = iu[0:dot_found]
            after_dot = iu[dot_found:]
            if after_dot.startswith('.Arn'):
                self.Integration.Uri = Sub(f'{before_dot}{stage}{after_dot}')


class ApiGatewayStage(agw.Stage):
    def __init__(self, title, name, key, **kwargs):
        super().__init__(title, **kwargs)
        self.StageName = name

        auto_get_props(self)

        self.RestApiId = Ref('ApiGatewayRestApi')
        self.DeploymentId = Ref(f'ApiGatewayDeployment{name}')
        self.DeploymentId = If(
            f'Deployment{name}A',
            Ref(f'ApiGatewayDeployment{name}A'),
            Ref(f'ApiGatewayDeployment{name}B'),
        )


class ApiGatewayDeployment(agw.Deployment):
    def __init__(self, title, name, key, **kwargs):
        super().__init__(title, **kwargs)
        lastresource = next(reversed(list(cfg.ApiGatewayResource)))
        lastmethod = (next(
            reversed(list(cfg.ApiGatewayResource[lastresource]['Method']))))
        self.DependsOn = f'ApiGatewayResource{lastresource}Method{lastmethod}'
        self.Description = Ref(f'Deployment{name}Description')
        self.RestApiId = Ref('ApiGatewayRestApi')


def AGW_Account(key):
    # Resources
    R_Account = ApiGatewayAccount('ApiGatewayAccount')

    add_obj([
        R_Account])


def AGW_DomainName(key):
    for n, v in getattr(cfg, key).items():
        if not v['IBOX_ENABLED']:
            continue

        resname = f'{key}{n}'
        # resources
        r_Domain = agw.DomainName(resname)
        auto_get_props(r_Domain)

        r_r53 = r53.RecordSetType(resname)
        if 'REGIONAL' in v['EndpointConfiguration']['Types']:
            auto_get_props(r_r53, 'R53RecordSetApiGatewayDomainNameRegional')
        else:
            auto_get_props(r_r53, 'R53RecordSetApiGatewayDomainNameGlobal')
        r_r53.title = f'RecordSet{resname}'

        # outputs
        o_Domain = Output(resname)
        o_Domain.Value = Ref(resname)
        o_Domain.Export = Export(resname)

        add_obj([
            r_Domain,
            r_r53,
            o_Domain])


def AGW_UsagePlans(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        for m, w in v['ApiStages'].items():
            # parameters
            p_Stage = Parameter(f'{resname}ApiStages{m}Stage')
            p_Stage.Description = (
                f'{m} Stage - empty for default based on env/role')

            add_obj(p_Stage)

        # resources
        r_UsagePlan = agw.UsagePlan(resname)
        auto_get_props(r_UsagePlan)

        # outputs
        o_UsagePlan = Output(resname)
        o_UsagePlan.Value = Ref(resname)
        o_UsagePlan.Export = Export(resname)

        add_obj([
            r_UsagePlan,
            o_UsagePlan])


def AGW_ApiKeys(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        # parameters
        p_Enabled = Parameter(f'{resname}Enabled')
        p_Enabled.Description = (
            f'{resname}Enabled - empty for mapped value')
        p_Enabled.AllowedValues = ['', 'yes', 'no']

        p_UsagePlan = Parameter(f'{resname}UsagePlan')
        p_UsagePlan.Description = (
            f'{resname}UsagePlan - empty for default based on env/role')

        add_obj([
            p_Enabled,
            p_UsagePlan])

        # resources
        r_ApiKey = agw.ApiKey(resname)
        auto_get_props(r_ApiKey)

        if 'UsagePlan' in v:
            plankey_name = f'{resname}UsagePlan'
            r_UsagePlanKey = agw.UsagePlanKey(f'ApiGatewayUsagePlan{n}')
            r_UsagePlanKey.KeyId = Ref(resname)
            r_UsagePlanKey.KeyType = 'API_KEY'
            r_UsagePlanKey.UsagePlanId = ImportValue(
                get_subvalue(
                    'ApiGatewayUsagePlan${1M}', f'{resname}UsagePlan')
            )

            add_obj(r_UsagePlanKey)

        # outputs
        o_ApiKey = Output(resname)
        o_ApiKey.Value = Ref(resname)

        add_obj([
            r_ApiKey,
            o_ApiKey])


def AGW_Stages(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        depname = f'Deployment{n}'
        # parameters
        p_DeploymentDescription = Parameter(f'{depname}Description')
        p_DeploymentDescription.Description = f'{depname} Description'
        p_DeploymentDescription.Default = n

        p_Deployment = Parameter(depname)
        p_Deployment.Description = (
            f'{depname} - change between A/B '
            'to trigger new deploy')
        p_Deployment.AllowedValues = ['A', 'B']
        p_Deployment.Default = 'A'

        add_obj([
            p_DeploymentDescription,
            p_Deployment])

        # conditons
        c_DeploymentA = get_condition(
            f'{depname}A', 'equals', 'A', depname, nomap=True)

        c_DeploymentB = get_condition(
            f'{depname}B', 'equals', 'B', depname, nomap=True)

        add_obj([
            c_DeploymentA,
            c_DeploymentB])

        # resources
        r_Stage = ApiGatewayStage(resname, name=n, key=v)

        r_DeploymentA = ApiGatewayDeployment(f'ApiGatewayDeployment{n}A',
                                             name=n, key=v)
        r_DeploymentA.Condition = f'{depname}A'

        r_DeploymentB = ApiGatewayDeployment(f'ApiGatewayDeployment{n}B',
                                             name=n, key=v)
        r_DeploymentB.Condition = f'{depname}B'

        # output
        o_Deployment = Output(depname)
        o_Deployment.Value = Ref(depname)

        add_obj([
            r_Stage,
            r_DeploymentA,
            r_DeploymentB,
            o_Deployment])


def AGW_RestApi(key):
    # Resources
    R_RestApi = agw.RestApi('ApiGatewayRestApi')
    auto_get_props(R_RestApi, f'{key}Base')
    R_RestApi.Policy = IAMPolicyApiGatewayPrivate()

    try:
        condition = cfg.PolicyCondition
        R_RestApi.Policy['Statement'][0]['Condition'] = condition
    except Exception:
        pass

    add_obj([
        R_RestApi,
    ])

    for n, v in cfg.ApiGatewayResource.items():
        resname = f'ApiGatewayResource{n}'
        agw_stage = cfg.ApiGatewayRestApi['Base']['Stage']
        r_Resource = ApiGatewayResource(resname, key=v, stage=agw_stage)

        for m, w in v['Method'].items():
            r_Method = ApiGatewayMethod(f'{resname}Method{m}',
                                        key=w, basename=n,
                                        name=m, stage=agw_stage)

            add_obj([
                r_Resource,
                r_Method])

    for n, v in cfg.Lambda.items():
        r_LambdaPermission = LambdaPermissionApiGateway(
            f'LambdaPermission{n}', name=Ref(f'Lambda{n}'),
            source=Sub('arn:aws:execute-api:${AWS::Region}:'
                       '${AWS::AccountId}:${ApiGatewayRestApi}/*/*/*')
        )

        add_obj(r_LambdaPermission)



import troposphere.cloudformation as cfm

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj, camel_to_snake)


def cfn_ecs_cluster():
    init_args = {}

    init_args.update({
        'REPOSITORIES': cfm.InitConfig(
            commands=If('GPUInstance', {
                '10-epel-repo': {
                    'command': 'amazon-linux-extras install -y epel',
                    'test': '! test -e /etc/yum.repos.d/epel.repo'
                },
                # 'cuda-repo': {
                #    'command': 'yum install -y '
                #    'http://developer.download.nvidia.com'
                #    '/compute/cuda/repos/rhel7/x86_64/'
                #    'cuda-repo-rhel7-10.0.130-1.x86_64.rpm',
                #    'test': 'test ! -e /etc/yum.repos.d/cuda.repo'
                # },
                'nvidia-docker': {
                    'command': 'curl -s -L '
                    'https://nvidia.github.io'
                    '/nvidia-docker/amzn2/nvidia-docker.repo | '
                    'tee /etc/yum.repos.d/nvidia-docker.repo',
                    'test': 'test ! -e /etc/yum.repos.d/nvidia-docker.repo'
                },
            }, Ref('AWS::NoValue')),
            files=If('GPUInstance', {
                '/etc/yum.repos.d/epel-nvidia.repo': {
                    'content': Join('\n', [
                        '[epel-nvidia]',
                        'name=negativo17 - Nvidia',
                        'baseurl=https://negativo17.org'
                        '/repos/nvidia/epel-7/$basearch/',
                        'enabled=1',
                        'skip_if_unavailable=1',
                        'gpgcheck=1',
                        'gpgkey=https://negativo17.org'
                        '/repos/RPM-GPG-KEY-slaanesh',
                        'enabled_metadata=1',
                        'metadata_expire=6h',
                        'type=rpm-md',
                        'repo_gpgcheck=0'
                    ])
                }
            }, Ref('AWS::NoValue'))
        ),
        'PACKAGES': cfm.InitConfig(
            packages={
                'yum': {
                    'nfs-utils': [],
                    'make': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'automake': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'gcc': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'gcc-c++': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'kernel-devel': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    # 'cuda': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'dkms-nvidia': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'nvidia-driver-cuda': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                    'nvidia-docker2': If(
                        'GPUInstance', [], Ref('AWS::NoValue')),
                }
            }
        ),
        'SERVICES': cfm.InitConfig(
            files={
                '/etc/ecs/ecs.config': {
                    'content': Join('\n', [
                        Sub('ECS_CLUSTER=${Cluster}')
                    ] + [
                        get_subvalue('%s=${1M}' % camel_to_snake(n),
                                     f'ECSClusterBaseAgentCfg{n}') for n in
                        cfg.ECSClusterBaseAgentCfg
                    ] + [
                        If(
                            'GPUInstance',
                            'ECS_DISABLE_PRIVILEGED=false',
                            Ref('AWS::NoValue')
                        ),
                    ])
                },
                '/etc/docker/daemon.json': If('GPUInstance', {
                    'content': Join('\n', [
                        '{',
                        '  "default-runtime": "nvidia",',
                        '  "runtimes": {',
                        '    "nvidia": {',
                        '      "path": "/usr/bin/nvidia-container-runtime",',
                        '      "runtimeArgs": []',
                        '    }',
                        '  }',
                        '}',
                    ])
                }, Ref('AWS::NoValue'))
            },
            commands={
                '01-kern-modules': If('GPUInstance', {
                    'command': 'rmmod nvidia_modeset;rmmod nvidia_uvm;'
                    'rmmod nvidia;modprobe nvidia;modprobe nvidia_uvm;true'
                }, Ref('AWS::NoValue')),
                '02-restart-docker': If('GPUInstance', {
                    'command': 'pkill -SIGHUP dockerd'
                }, Ref('AWS::NoValue')),
            }
        )
    })

    # aws ecs ami by default run postfix, use yaml cfg Postfix: false to
    # stop it
    try:
        cfg.Postfix
    except Exception:
        pass
    else:
        # Parameter
        P_Postfix = Parameter('Postfix')
        P_Postfix.Description = 'Disable Postfix Service'
        # Condition
        C_Postfix = get_condition('Postfix', 'equals', 'no')
        # Output
        O_Postfix = Output('Postfix')
        O_Postfix.Value = get_endvalue('Postfix')

        add_obj([
            P_Postfix,
            C_Postfix,
            O_Postfix,
        ])

        init_args['SERVICES'].services = {
            'sysvinit': {
                'postfix': If(
                    'Postfix',
                    {
                        'enabled': 'false',
                        'ensureRunning': 'false'},
                    Ref('AWS::NoValue'))}}

    return init_args



import troposphere.autoscaling as asg
import troposphere.ec2 as ec2
import troposphere.cloudformation as cfm
import troposphere.policies as pol
import troposphere.applicationautoscaling as aas

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .cfn import *
from .codedeploy import CD_DeploymentGroup
from .iam import IAMInstanceProfile

try:
    from cfnExt import *
except ModuleNotFoundError:
    pass


class ASLaunchTemplateData(ec2.LaunchTemplateData):
    def __init__(self, title, UserDataApp, **kwargs):
        super().__init__(title, **kwargs)
        auto_get_props(self, 'LaunchTemplateData')
        user_data = [
            '#!/bin/bash',
            'PATH=/opt/aws/bin:$PATH',
            'export BASH_ENV=/etc/profile.d/ibox_env.sh',
            'export ENV=$BASH_ENV',
            'yum -C list installed aws-cfn-bootstrap ||'
            ' yum install -y aws-cfn-bootstrap',
            Sub(''.join(UserDataApp)),
            Sub(
                'cfn-init -v --stack ${AWS::StackName} --role ${RoleInstance}'
                ' --resource LaunchTemplate --region ${AWS::Region}'),
            If(
                'DoNotSignal',
                Ref('AWS::NoValue'),
                Sub(
                    'cfn-signal -e $? --stack ${AWS::StackName}'
                    ' --role ${RoleInstance} --resource AutoScalingGroup'
                    ' --region ${AWS::Region}')
            ),
            'rm /var/lib/cloud/instance/sem/config_scripts_user\n',
        ]

        try:
            cfg.BottleRocket
        except Exception:
            self.UserData = Base64(Join('\n', user_data))
        else:
            self.UserData = Base64(Join('\n', If(
                'BottleRocket',
                [
                    get_endvalue(f'BottleRocketUserData{n}Line')
                    for n in cfg.BottleRocketUserData] + ['\n'],
                user_data,
            )))


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class ASInitConfigSets(cfm.InitConfigSets):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if cfg.Apps:
            CODEDEPLOY = If(
                'DeploymentGroup', 'CODEDEPLOY', Ref('AWS::NoValue'))
        else:
            CODEDEPLOY = Ref('AWS::NoValue')

        CWAGENT = If('CloudWatchAgent', 'CWAGENT', Ref('AWS::NoValue'))

        if getattr(cfg, 'IBOX_LAUNCH_TEMPLATE_NO_WAIT_ELB_HEALTH', False):
            ELBWAITER = Ref('AWS::NoValue')
        elif cfg.LoadBalancerClassic or cfg.LoadBalancerApplication:
            ELBWAITER = 'ELBWAITER'
        else:
            ELBWAITER = Ref('AWS::NoValue')

        self.data = {
            'default': [
                'REPOSITORIES',
                'PACKAGES',
                'SETUP',
                CODEDEPLOY,
                'SERVICES',
                CWAGENT,
                ELBWAITER,
            ],
            'buildami': [
                'REPOSITORIES',
                'PACKAGES',
            ],
            'buildamifull': [
                'REPOSITORIES',
                'PACKAGES',
                'SETUP',
            ]
        }


class ASInitConfigSetup(cfm.InitConfig):
    def __init__(self, **kwargs):
        self.ibox_env_app = ''
        super(ASInitConfigSetup, self).__init__(**kwargs)

    def setup(self):
        self.files = {
            '/etc/profile.d/ibox_env.sh': {
                'content': Join('', [
                    '#setup ibox environment variables\n',
                    'export Env=', Ref('Env'), '\n',
                    'export EnvAbbr=', Ref('EnvShort'), '\n',
                    'export EnvRegion=', Ref('AWS::Region'), '\n',
                    'export EnvAccountId=', Ref('AWS::AccountId'), '\n',
                    'export EnvRole=', Ref('EnvRole'), '\n',
                    'export EnvBrand=', cfg.BrandDomain, '\n',
                    'export EnvStackName=', Ref('AWS::StackName'), '\n',
                ] + self.ibox_env_app)
            },
            '/etc/cfn/cfn-hup.conf': {
                'content': Join('', [
                    '[main]\n',
                    'stack=', Ref('AWS::StackId'), '\n',
                    'region=', Ref('AWS::Region'), '\n',
                    'role=', Ref('RoleInstance'), '\n',
                    'interval=5\n'
                ]),
                'mode': '000400',
                'owner': 'root',
                'group': 'root'
            },
            '/etc/cfn/hooks.d/cfn-auto-reloader.conf': {
                'content': Join('', [
                    '[cfn-auto-reloader-hook]\n',
                    'triggers=post.add, post.update\n',
                    'path=Resources.LaunchTemplate'
                    '.Metadata.CloudFormationInitVersion\n',
                    'action=/opt/aws/bin/cfn-init -v',
                    ' --stack ', Ref('AWS::StackName'),
                    ' --role ', Ref('RoleInstance'),
                    ' --resource LaunchTemplate',
                    ' --region ', Ref('AWS::Region'), '\n',
                    'runas=root\n'
                ])
            },
            # '/usr/local/bin/chamber': {
            #     'mode': '000755',
            #     'source': Sub(
            #         f'https://{cfg.BucketAppRepository}'
            #         '.s3.${AWS::Region}.amazonaws.com/ibox-tools/chamber'),
            #     'owner': 'root',
            #     'group': 'root',
            # },
        }
        self.commands = {
            '01_disk': {
                'command': Join('', [
                    'n=0\n',
                    'for disk in /dev/xvd[b-d]; do\n',
                    '  [ -b "$disk" ] || continue\n',
                    '  file -s "$disk" | grep -q "ext[34] filesystem" || ',
                    '  { mkfs.ext4 $disk || continue; }\n',
                    '  mkdir -p /media/ephemeral${n} && ',
                    '  mount $disk /media/ephemeral${n}\n',
                    '  n=$(($n+1))\n',
                    'done'
                ])
            },
            '02_disk_additional': If(
                'LaunchTemplateDataBlockDeviceMappingsAdditionalStorageMount',
                {
                    'command': get_subvalue(
                        'file -s ${1M}1 | grep -q "ext[34] filesystem" ||'
                        ' { parted -s ${1M} mklabel gpt &&'
                        ' parted -s ${1M} mkpart primary ext2 1 ${2M}G &&'
                        ' mkfs.ext4 ${1M}1 || continue; }\nmkdir -p /data &&'
                        ' mount ${1M}1 /data',
                        [
                            ('LaunchTemplateDataBlockDeviceMappings'
                             'AdditionalStorageDeviceName'),
                            ('LaunchTemplateDataBlockDeviceMappings'
                             'AdditionalStorageEbsVolumeSize'),
                        ]
                    )
                },
                Ref('AWS::NoValue')
            ),
            '03_efs_mounts': If(
                'EfsMounts',
                {
                    'command': Join('', [
                        'for mounts in ', Join(' ', Ref('EfsMounts')), ';do\n',
                        '  mkdir -p "/media/${mounts}"\n',
                        '  mountpoint -q "/media/${mounts}" ||',
                        '    mount -t nfs4 -o nfsvers=4,rsize=1048576,'
                        'wsize=1048576,hard,timeo=600,retrans=2 ',
                        '    efs-${mounts}.', cfg.HostedZoneNamePrivate, ':/ ',
                        '    /media/${mounts}\n',
                        'done'
                    ]),
                },
                Ref('AWS::NoValue')
            ),
            '04_rmdir_tmp_ibox': {
                'command': 'rm -fr /tmp/ibox'
            }
        }
        self.services = {
            'sysvinit': {
                'cfn-hup': {
                    'enabled': 'false',
                    'ensureRunning': 'true',
                    'files': [
                        '/etc/cfn/cfn-hup.conf',
                        '/etc/cfn/hooks.d/cfn-auto-reloader.conf'
                    ]
                }
            }
        }


class ASInitConfigCodeDeploy(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.files = {
            '/etc/codedeploy-agent/conf/codedeployagent.yml': {
                'content': Join('', [
                    '---\n',
                    ':log_aws_wire: false\n',
                    ':log_dir: \'/var/log/aws/codedeploy-agent/\'\n',
                    ':pid_dir: \'/opt/codedeploy-agent/state/.pid/\'\n',
                    ':program_name: codedeploy-agent\n',
                    ':root_dir: \'/opt/codedeploy-agent/deployment-root\'\n',
                    ':verbose: false\n',
                    ':wait_between_runs: 1\n',
                    ':proxy_uri:\n',
                    ':max_revisions: 2\n'
                ])
            },
            '/tmp/codedeployinstall.sh': {
                'source': Sub(
                    'https://aws-codedeploy-${AWS::Region}'
                    '.s3.amazonaws.com/latest/install'),
                'mode': '000700'
            }
        }
        self.commands = {
            '03_codedeploy-install-and-run': {
                'command': '/tmp/codedeployinstall.sh auto'
            }
        }
        self.services = {
            'sysvinit': {
                'codedeploy-agent': {
                    'enabled': 'false',
                    'ensureRunning': 'true',
                    'files': [
                        '/etc/codedeploy-agent/conf/codedeployagent.yml'
                    ]
                }
            }
        }


class ASInitConfigCloudWatchAgent(cfm.InitConfig):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        self.packages = {
            'rpm': {
                'amazon-cloudwatch-agent': (
                    'https://s3.amazonaws.com'
                    '/amazoncloudwatch-agent/amazon_linux/amd64/latest/'
                    'amazon-cloudwatch-agent.rpm'
                )
            }
        }
        self.files = {
            '/opt/aws/amazon-cloudwatch-agent/etc/'
            'amazon-cloudwatch-agent.json': {
                'content': Join('', [
                    '{\n',
                    '  "metrics": {\n',
                    '    "append_dimensions": {\n',
                    '      "AutoScalingGroupName": '
                    '"${aws:AutoScalingGroupName}",\n',
                    '      "InstanceId": "${!aws:InstanceId}"\n',
                    '    },\n',
                    '    "aggregation_dimensions": [\n',
                    '      ["AutoScalingGroupName"]\n',
                    '    ],\n',
                    '    "metrics_collected": {\n',
                    '      "mem": {\n',
                    '        "measurement": [\n',
                    '          "mem_used_percent"\n',
                    '        ]\n',
                    '      },\n',
                    '      "disk": {\n',
                    '        "resources": [\n',
                    '          "/"\n',
                    '        ],\n',
                    '        "measurement": [\n',
                    '          {\n',
                    '            "name": "disk_used_percent",\n',
                    '            "rename": "root_disk_used_percent"\n',
                    '          }\n',
                    '        ],\n',
                    '        "drop_device": true\n',
                    '      }\n',
                    '    }\n',
                    '  }\n',
                    '}\n',
                ]),
            },
        }
        self.services = {
            'sysvinit': {
                'amazon-cloudwatch-agent': {
                    'ensureRunning': 'true',
                    'files': [
                        '/opt/aws/amazon-cloudwatch-agent/etc/'
                        'amazon-cloudwatch-agent.json'
                    ]
                }
            }
        }


class ASInitConfigApps(cfm.InitConfig):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. Apps1
        reponame = f'{name}RepoName'
        n = name.replace('Apps', '')
        envappversion = f'EnvApp{n}Version'

        self.sources = {
            '/tmp/ibox/': If(
                'DeployRevision',
                Ref('AWS::NoValue'),
                get_subvalue(
                    'https://%s.s3-${AWS::Region}.amazonaws.com/'
                    '${1M}/${1M}-${%s}.tar.gz' % (
                        cfg.BucketAppRepository, envappversion),
                    reponame, ''
                )
            )
        }
        self.commands = {
            '01_setup': If(
                'DeployRevision',
                Ref('AWS::NoValue'),
                {
                    'command': get_subvalue(
                        'EnvAppVersion=${%s} EnvRepoName=${1M} '
                        '/tmp/ibox/bin/setup.sh' % envappversion,
                        reponame
                    )
                }
            ),
            '02_setup_reboot_codedeploy': If(
                'DeployRevision',
                {
                    'command': get_subvalue(
                        'EnvAppVersion=${%s} EnvRepoName=${1M} '
                        '/opt/ibox/${1M}/live/bin/setup.sh' % envappversion,
                        reponame
                    ),
                    'test': 'test -e /var/lib/cloud/instance/sem/'
                    'config_ssh_authkey_fingerprints'
                },
                Ref('AWS::NoValue')
            ),
            '03_rmdir_tmp_ibox': {
                'command': 'rm -fr /tmp/ibox'
            }
        }


class ASInitConfigAppsBuildAmi(ASInitConfigApps):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        for n, v in self.sources.items():
            if not isinstance(v, dict):
                if 'Fn::If' in v.data:
                    self.sources[n] = v.data['Fn::If'][2]

        for n, v in self.commands.items():
            if not isinstance(v, dict):
                if 'Fn::If' in v.data:
                    self.commands[n] = v.data['Fn::If'][2]


class ASInitConfigELBClassicExternal(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            'ELBClassicExternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"InService"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'),
                    ' elb describe-instance-health',
                    '  --load-balancer-name ',
                    Ref('LoadBalancerClassicExternal'),
                    '  --instances $(curl -s '
                    'http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query InstanceStates[0].State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBClassicInternal(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            'ELBClassicInternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"InService"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'),
                    ' elb describe-instance-health',
                    '  --load-balancer-name ',
                    Ref('LoadBalancerClassicInternal'),
                    '  --instances $(curl -s '
                    'http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query InstanceStates[0].State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBApplicationExternal(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            'ELBApplicationExternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"healthy"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'),
                    ' elbv2 describe-target-health',
                    '  --target-group-arn ', Ref('TargetGroupExternal'),
                    '  --targets Id=$(curl -s '
                    'http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query '
                    'TargetHealthDescriptions[0].TargetHealth.State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBApplicationInternal(cfm.InitConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.commands = {
            'ELBApplicationInternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"healthy"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'),
                    ' elbv2 describe-target-health',
                    '  --target-group-arn ', Ref('TargetGroupInternal'),
                    '  --targets Id=$(curl -s '
                    'http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query '
                    'TargetHealthDescriptions[0].TargetHealth.State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

def AS_ScalingPolicies(key):
    Out_String = []
    Out_Map = {}
    for n, v in getattr(cfg, key).items():
        if not v['IBOX_ENABLED']:
            continue

        resname = f'{key}{n}'

        # resources
        if key == 'AutoScalingScalingPolicy':
            r_Policy = asg.ScalingPolicy(resname)
        else:
            r_Policy = aas.ScalingPolicy(resname)

        auto_get_props(r_Policy)
        add_obj(r_Policy)

        # for tracking create output
        if v['PolicyType'] == 'TargetTrackingScaling':
            # Autoscaling
            if 'TargetTrackingConfiguration' in v:
                TargetTrackingConfigurationName = (
                    'TargetTrackingConfiguration')
            # Application Autoscaling
            elif 'TargetTrackingScalingPolicyConfiguration' in v:
                TargetTrackingConfigurationName = (
                    'TargetTrackingScalingPolicyConfiguration')

            basename = f'{resname}{TargetTrackingConfigurationName}'

            # outputs
            if v['Type'] == 'Cpu' or (
                    v['Type'] == 'Custom' and
                    v[TargetTrackingConfigurationName]
                    ['CustomizedMetricSpecification']
                    ['MetricName'] == 'CPUUtilization'):
                # Use Cpu Metric
                Out_String.append('Cpu${Statistic}:${Cpu}')

                if v['Type'] == 'Custom':
                    statistic = get_endvalue(
                        f'{basename}'
                        'CustomizedMetricSpecificationStatistic')
                else:
                    statistic = ''

                Out_Map.update({
                    'Statistic': statistic,
                    'Cpu': get_endvalue(f'{basename}TargetValue'),
                })

    if Out_String:
        # Outputs
        O_Policy = Output(key)
        O_Policy.Value = Sub(','.join(Out_String), **Out_Map)

        add_obj(O_Policy)


def AS_LaunchTemplate():
    InitConfigSets = ASInitConfigSets()

    CfmInitArgs = {}
    IBoxEnvApp = []
    Tags_List = []
    UserDataApp = []

    for n in cfg.Apps:
        name = f'Apps{n}'              # Ex. Apps1
        envname = f'EnvApp{n}Version'  # Ex EnvApp1Version
        reponame = f'{name}RepoName'   # Ex Apps1RepoName

        UserDataApp.extend(['#${%s}\n' % envname])

        p_EnvAppVersion = Parameter(envname)
        p_EnvAppVersion.Description = f'Application {n} version'
        p_EnvAppVersion.AllowedPattern = '^[a-zA-Z0-9-_.]*$'

        p_AppsRepoName = Parameter(reponame)
        p_AppsRepoName.Description = (
            f'App {n} Repo Name - empty for default based on env/role')
        p_AppsRepoName.AllowedPattern = '^[a-zA-Z0-9-_.]*$'

        # parameters
        add_obj([
            p_EnvAppVersion,
            p_AppsRepoName,
        ])

        # conditions
        add_obj({
            name: And(
                Not(Equals(Ref(envname), '')),
                Not(get_condition('', 'equals', 'None', reponame)))})

        InitConfigApps = ASInitConfigApps(name)
        CfmInitArgs[name] = InitConfigApps

        InitConfigAppsBuilAmi = ASInitConfigAppsBuildAmi(name)
        # AUTOSPOT - Let cfn-init always prepare instances on boot
        # CfmInitArgs[name + 'BuildAmi'] = InitConfigAppsBuilAmi
        CfmInitArgs[name] = InitConfigAppsBuilAmi

        IBoxEnvApp.extend([
            f'export EnvApp{n}Version=', Ref(envname), "\n",
            f'export EnvRepo{n}Name=', get_endvalue(reponame), "\n"])

        InitConfigSetsApp = If(name, name, Ref('AWS::NoValue'))
        InitConfigSetsAppBuilAmi = If(
            name, f'{name}BuildAmi', Ref('AWS::NoValue'))
        IndexSERVICES = InitConfigSets.data['default'].index('SERVICES')
        InitConfigSets.data['default'].insert(
            IndexSERVICES, InitConfigSetsApp)
        # AUTOSPOT - Let cfn-init always prepare instances on boot
        # InitConfigSets.data['buildamifull'].append(
        #    InitConfigSetsAppBuilAmi)
        InitConfigSets.data['buildamifull'].append(InitConfigSetsApp)

        Tags_List.append(asg.Tag(envname, Ref(envname), True))

        # resources
        # FOR MULTIAPP CODEDEPLOY
        if len(cfg.Apps) > 1:
            r_DeploymentGroup = CDDeploymentGroup(f'DeploymentGroup{name}')
            r_DeploymentGroup.setup(index=n)

            add_obj(r_DeploymentGroup)

        # outputs
        Output_app = Output(envname)
        Output_app.Value = Ref(envname)
        add_obj(Output_app)

        Output_repo = Output(reponame)
        Output_repo.Value = get_endvalue(reponame)
        add_obj(Output_repo)

    InitConfigSetup = ASInitConfigSetup()
    InitConfigSetup.ibox_env_app = IBoxEnvApp
    InitConfigSetup.setup()

    InitConfigCodeDeploy = ASInitConfigCodeDeploy()

    CfmInitArgs['SETUP'] = InitConfigSetup
    CfmInitArgs['CWAGENT'] = ASInitConfigCloudWatchAgent('')

    if cfg.Apps:
        CfmInitArgs['CODEDEPLOY'] = InitConfigCodeDeploy
        CD_DeploymentGroup()

    if not getattr(cfg, 'IBOX_LAUNCH_TEMPLATE_NO_WAIT_ELB_HEALTH', False):
        # LoadBalancerClassic External
        if cfg.LoadBalancerClassicExternal:
            InitConfigELBExternal = ASInitConfigELBClassicExternal()
            CfmInitArgs['ELBWAITER'] = InitConfigELBExternal

        # LoadBalancerClassic Internal
        if cfg.LoadBalancerClassicInternal:
            InitConfigELBInternal = ASInitConfigELBClassicInternal()
            CfmInitArgs['ELBWAITER'] = InitConfigELBInternal

        # LoadBalancerApplication External
        if cfg.LoadBalancerApplicationExternal:
            InitConfigELBExternal = ASInitConfigELBApplicationExternal()
            CfmInitArgs['ELBWAITER'] = InitConfigELBExternal

        # LoadBalancerApplication Internal
        if cfg.LoadBalancerApplicationInternal:
            InitConfigELBInternal = ASInitConfigELBApplicationInternal()
            CfmInitArgs['ELBWAITER'] = InitConfigELBInternal

    if getattr(cfg, 'IBOX_LAUNCH_TEMPLATE_NO_SG_EXTRA', False):
        SecurityGroups = []
    else:
        SecurityGroups = cfg.SecurityGroupsImport

    # Resources
    R_LaunchTemplate = ec2.LaunchTemplate(
        'LaunchTemplate',
        LaunchTemplateName=Sub('${AWS::StackName}-${EnvRole}'),
        LaunchTemplateData=ASLaunchTemplateData('LaunchTemplateData',
                                                UserDataApp=UserDataApp))
    R_LaunchTemplate.LaunchTemplateData.NetworkInterfaces[0].Groups.extend(
        SecurityGroups)

    R_InstanceProfile = IAMInstanceProfile('InstanceProfile')

    # Import role specific cfn definition
    try:
        # Do not use role but direct cfg yaml configuration (ecs + cluster)
        cfn_envrole = cfg.IBOX_CFN_FUNC
    except Exception:
        cfn_envrole = f'cfn_{cfg.func_envrole}'
    if cfn_envrole in globals():  # Ex cfn_client_portal
        CfnRole = globals()[cfn_envrole]()
        CfmInitArgs.update(CfnRole)

    R_LaunchTemplate.Metadata = cfm.Metadata(
        {
            'CloudFormationInitVersion': If(
                'CloudFormationInit',
                Ref('EnvStackVersion'),
                Ref('AWS::NoValue'),
            )
        },
        cfm.Init(
            InitConfigSets,
            **CfmInitArgs
        ),
        cfm.Authentication({
            'CfnS3Auth': cfm.AuthenticationBlock(
                type='S3',
                buckets=[
                    Sub(cfg.BucketAppRepository),
                    Sub(cfg.BucketAppData)
                ],
                roleName=Ref('RoleInstance')
            )
        })
    )

    add_obj([
        R_LaunchTemplate,
        R_InstanceProfile])

    Tags = asg.Tags()
    Tags.tags = Tags_List
    return Tags


def AS_Autoscaling(key):
    LoadBalancers = []
    for n in cfg.LoadBalancerClassic:
        LoadBalancers.append(Ref(f'LoadBalancerClassic{n}'))

    TargetGroups = []
    for n in cfg.LoadBalancerApplication:
        TargetGroups.append(Ref(f'TargetGroup{n}'))

    # Resources
    LaunchTemplateTags = AS_LaunchTemplate()

    R_ASG = asg.AutoScalingGroup('AutoScalingGroupBase')
    R_ASG.LoadBalancerNames = LoadBalancers
    R_ASG.TargetGroupARNs = TargetGroups

    auto_get_props(R_ASG)

    R_ASG.Tags += LaunchTemplateTags

    add_obj([
        R_ASG])


def AS_ScheduledActioneEC2(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        r_ScheduledActions = asg.ScheduledAction(resname)
        auto_get_props(r_ScheduledActions)
        add_obj(r_ScheduledActions)


def AS_ScalableTargetECS(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        # resources
        # trick - use fixed name ScalableTarget to avoid changin too much
        # code for now
        r_ScalableTarget = aas.ScalableTarget('ScalableTarget')
        auto_get_props(r_ScalableTarget, f'{key}Service')

        ScheduledActions = []
        for m, w in cfg.ScalableTargetScheduledAction.items():
            if not w.get('IBOX_ENABLED', True):
                continue
            sc_name = f'ScalableTargetScheduledAction{m}'
            r_ScheduledAction = aas.ScheduledAction(sc_name)
            auto_get_props(r_ScheduledAction)
            ScheduledActions.append(If(
                f'{sc_name}Disable',
                Ref('AWS::NoValue'),
                r_ScheduledAction))
        r_ScalableTarget.ScheduledActions = ScheduledActions

        add_obj([
            r_ScalableTarget])



import yaml
import yaml.constructor
import sys
import os
import copy
import json
import logging
from pprint import pprint, pformat

from . import cfg


def build_RP():
    LD_INCLUDED = []
    LD_EXCLUDED = []

    CFG_FILE_INT = '%s/cfg' % os.path.dirname(os.path.realpath(__file__))
    CFG_FILE_INT = os.path.normpath(CFG_FILE_INT)

    CFG_FILE_EXT = '%s/cfg' % os.getcwd()
    CFG_FILE_EXT = os.path.normpath(CFG_FILE_EXT)

    envrole = cfg.envrole
    stacktype = cfg.stacktype
    brand = cfg.brand

    RP_base = {
        'cmm': {
            'cmm': {}
        }
    }

    # dynamically populate RP_base Dict from ENV_BASE and cfg.regions
    for n in cfg.ENV_BASE:
        RP_base[n] = {}
        for m in cfg.regions:
            RP_base[n][m] = {}

    cfg.RP_base = RP_base

    def _get_RP_base_keys():
        RP_base_keys = ['global']
        for n, v in RP_base.items():
            RP_base_keys.append(n)
            for m, w in v.items():
                RP_base_keys.append(m)

        return RP_base_keys + [stacktype, envrole]

    RP_base_keys = _get_RP_base_keys()

    class Loader(yaml.Loader):
        def __init__(self, stream):
            # This way for include relative to file with include statement
            self._root_current = os.path.split(stream.name)[0]
            # This way for include BASE relative on BASE dir
            self._root_base = os.path.join(CFG_FILE_INT, 'BASE')
            # This way for include relative on BASE EXT dir
            self._root_base_ext = os.path.join(CFG_FILE_EXT, 'BASE')
            # This way for include relative on brand EXT dir
            self._root_brand_ext = os.path.join(CFG_FILE_EXT, brand)

            # try to find out if source file is in a subfolder
            if self._root_base in self._root_current:
                suffix = self._root_current.replace(self._root_base, '')
            elif self._root_base_ext in self._root_current:
                suffix = self._root_current.replace(self._root_base_ext, '')
            else:
                suffix = self._root_current.replace(self._root_brand_ext, '')

            # self.root_current_suffix will be empty if source file is in root
            self.root_current_suffix = os.path.basename(suffix)
            if self.root_current_suffix.upper() in cfg.STACK_TYPES:
                # set to empty even if UPPERCASE subfolders is in
                # cfg.STACK_TYPES to not be searched for include files
                self.root_current_suffix = ''

            self.stream = stream
            super(Loader, self).__init__(stream)
            Loader.add_constructor('!include', Loader.include)
            Loader.add_constructor('!exclude', Loader.exclude)

        def exclude(self, node):
            if isinstance(node, yaml.SequenceNode):
                for filename in self.construct_sequence(node):
                    LD_EXCLUDED.append(filename)

        def include(self, node):
            if isinstance(node, yaml.SequenceNode):
                result = []
                for filename in self.construct_sequence(node):
                    if filename in LD_EXCLUDED + LD_INCLUDED:
                        # a yml file can be included only once
                        continue
                    LD_INCLUDED.append(filename)

                    # search for include in files in roots
                    for path in [self._root_base,
                                 self._root_base_ext,
                                 self._root_brand_ext]:
                        contents = self.extractFile(filename, path)
                        if contents:
                            result.append(contents)
                        if self.root_current_suffix:
                            # search in subfolder too
                            contents = self.extractFile(filename, os.path.join(
                                path, self.root_current_suffix))
                            if contents:
                                result.append(contents)
                return result
            else:
                print('Error:: unrecognised node type in !include statement')
                raise yaml.constructor.ConstructorError

        def extractFile(self, filename, root):
            filepath = os.path.join(root, filename)
            try:
                with open(filepath, 'r') as f:
                    return yaml.load(f, Loader)
            except IOError:
                pass

    def process_cfg(cfg, envs, check_root=None):
        if hasattr(cfg, 'items'):
            # This method allow to delete items from a dictionary
            # while iterating over it
            for k in list(cfg):
                v = cfg[k]
                # for final values
                if (isinstance(v, (str, int, list))
                        and not k.startswith('IBoxLoader')):
                    yield {k: v}
                # for empty dict in common.yml
                if isinstance(v, dict) and len(v) == 0:
                    yield {k: v}
                # for recursively descending in env/region role dict.
                # list is needed for IBoxLoader include list.
                if k in envs and isinstance(v, (dict, list)):
                    if k in ['IBoxLoader', 'IBoxLoaderAfter']:
                        # IBoxLoader included dict processed by
                        # process_cfg have check_root = True
                        kwargs = {'check_root': True}
                    else:
                        kwargs = {}
                    try:
                        # after descending in env main key
                        # (not the one nested under region) delete key
                        # this way when envs include both (env and region)
                        # i do not process it again
                        if k in list(RP_base.keys()):
                            del cfg[k]
                    except Exception:
                        pass
                    for result in process_cfg(v, envs, **kwargs):
                        yield result
                if check_root:
                    # Here i do not have yet encountered a envs key in an
                    # IBoxLoader included dict, skip properties until i find it
                    continue
                # for recursively descending in dict not in RP_base_keys
                # (env/region/envrole/stacktype)
                # (final key is the concatenation of traversed dict keys)
                if k not in RP_base_keys and isinstance(v, dict):
                    for j, w in v.items():
                        for result in process_cfg({f'{k}{j}': w}, envs):
                            yield result
        if isinstance(cfg, list):
            for n in cfg:
                for result in process_cfg(n, envs, check_root=check_root):
                    yield result

    def replace_not_allowed_char(s):
        # CF Mapping allow for index only alfanumeric char
        # this way i can specify more "clear" name
        # for index in CloudFormation behaviours
        key = str(s)
        for s, w in cfg.CLF_PATH_PATTERN_REPLACEMENT.items():
            key = key.replace(s, w)

        return int(key) if key.isdigit() else key

    def get_RP_for_envs(data):
        def _merge(base, work):
            if isinstance(work, (str, list)) or not base:
                return work
            keys = dict(list(base.items()) + list(work.items())).keys()
            for k in keys:
                if (isinstance(base.get(k), dict)
                        and isinstance(work.get(k), dict)):
                    base[k] = _merge(base[k], work[k])
                elif k.endswith('++') and isinstance(work.get(k), list):
                    # ++ is used to append elements to an existing key
                    base[k.replace('++', '')] += work[k]
                elif k in work:
                    base[k] = work[k]
            return base

        def _process(key, data, RP, merge=True):
            key = str(key)

            if key.startswith('/'):
                # for CFront Behaviors
                key = replace_not_allowed_char(key)
            if key.endswith('**'):
                # ** is used to replace existing dict instead of merging it
                key = key.replace('**', '')
                merge = False

            if merge and isinstance(RP.get(key), dict):
                # RP[key] already exist as a dict, try merging
                RP[key] = _merge(RP[key], _recurse(data))
            else:
                RP[key] = _recurse(data)

        def _recurse(data):
            RP = {}
            if isinstance(data, dict):
                # data is a dict
                for n, v in data.items():
                    _process(n, v, RP)
            elif isinstance(data, list) and data and isinstance(data[0], dict):
                # data is a list of dicts
                for v in data:
                    for m, w in v.items():
                        _process(m, w, RP)
            else:
                RP = data
            return RP

        return _recurse(data)

    def read_yaml(file_type, brand, base_dir, stacktype=''):
        cfg_file = os.path.join(
            base_dir, brand, stacktype.upper(), f'{file_type}.yml')

        try:
            with open(cfg_file, 'r') as ymlfile:
                cfg = yaml.load(ymlfile, Loader=Loader)

                return cfg
        except IOError:
            return {}

    def merge_cfg(cfgs, cfg_key, list_base=None):
        RP_list = copy.deepcopy(list_base) if list_base else []

        def _parse_cfg(cfg, envs=[]):
            parsed_cfg = {}
            for value in process_cfg(cfg, envs):
                for k, v in value.items():
                    try:
                        # if v is a list of dict and the same key (k) already
                        # exist, final value is the sum of the list (of dicts)
                        if isinstance(v[0], dict):
                            parsed_cfg[k] = parsed_cfg[k] + v
                        else:
                            raise
                    except Exception:
                        parsed_cfg[k] = v
            return parsed_cfg

        for cfg, v in cfgs.items():
            for c in v:
                keys = ['IBoxLoader', 'IBoxLoaderAfter'] + cfg_key[cfg]
                RP_list.append(_parse_cfg(c, keys))

        # RP_list is now a list of dict
        return RP_list

    def prepend_base_cfgs(cfg_cmm):
        base_cfgs = {}
        key_names = {}
        for c in cfg_cmm:
            if isinstance(c, dict):
                for n in list(c):
                    # need to delete items while iterating
                    v = c[n]
                    if n in cfg.BASE_CFGS:
                        if n not in base_cfgs:
                            base_cfgs[n] = {}
                        if n not in key_names:
                            key_names[n] = []
                        base_key_value = cfg.BASE_CFGS[n]
                        for k in list(v):
                            try:
                                base_value = k[base_key_value]
                            except Exception:
                                if isinstance(base_key_value, dict):
                                    # use dict value in cfg.BASE_CFGS
                                    base_cfgs[n] = base_key_value
                                key_names[n].extend(list(k.keys()))
                            else:
                                # use IBOX_BASE (or values in cfg.BASE_CFGS)
                                for j, w in base_value.items():
                                    j_current = base_cfgs[n].get(j)
                                    if j_current and isinstance(
                                            j_current, list):
                                        # if property is a list extend it
                                        base_cfgs[n][j].extend(w)
                                    else:
                                        base_cfgs[n][j] = w
                                v.remove(k)
                        if not c[n]:
                            del c[n]

        for n, v in key_names.items():
            values = []
            # is better "set(v):" but change order so for now keep it
            for m in v:
                values.append({m: base_cfgs[n]})
            if values:
                cfg_cmm.insert(0, {n: values})

    def get_RP(cfgs):
        RP = copy.deepcopy(RP_base)

        cfg_key_cmm = {
            'common': ['global'],
            'type': ['global'],
            'role': ['global'],
        }

        cfg_merge_cmm = merge_cfg(cfgs, cfg_key_cmm)

        # Prepend base config from awsibox/cfg.py BASE_CFGS
        prepend_base_cfgs(cfg_merge_cmm)

        RP['cmm']['cmm'] = get_RP_for_envs(cfg_merge_cmm)

        for env, rvalue in RP.items():
            if env == 'cmm':
                continue

            cfg_key_env = {
                'common': [env],
                'type': [env],
                'role': [env],
            }

            cfg_merge_env = merge_cfg(cfgs, cfg_key_env)

            for region in rvalue.keys():

                cfg_key_region = {
                    'common': [region],
                    'type': [region],
                    'role': [region],
                }

                cfg_key_env_region = {
                    'common': [env, region],
                    'type': [env, region],
                    'role': [env, region],
                }

                cfg_merge_region = merge_cfg(
                    cfgs, cfg_key_region, cfg_merge_env)
                cfg_merge_env_region = merge_cfg(
                    cfgs, cfg_key_env_region, cfg_merge_region)

                RP[env][region] = get_RP_for_envs(cfg_merge_env_region)

        return(RP)

    def set_cfg():
        def RP_to_cfg(key):
            if hasattr(key, 'items'):
                for k, v in key.items():
                    setattr(cfg, k, v)
                    # recursively traverse dict
                    # keys name are the concatenation of traversed dict keys
                    if isinstance(v, dict):
                        for j, w in v.items():
                            RP_to_cfg({f'{k}{j}': w})

        # for n, v in cfg.RP_cmm.items():
        #     setattr(cfg, n, v)
        RP_to_cfg(cfg.RP_cmm)

        # set generic attribute based on condition:

        # LoadBalancerClassic
        try:
            cfg.LoadBalancerClassic
        except Exception:
            cfg.LoadBalancerClassic = []

        # LoadBalancerApplication
        try:
            cfg.LoadBalancerApplication
        except Exception:
            cfg.LoadBalancerApplication = []

        # LoadBalancer
        cfg.LoadBalancer = None
        for n in ['LoadBalancerClassic', 'LoadBalancerApplication']:
            try:
                getattr(cfg, n)
            except Exception:
                pass
            else:
                cfg.LoadBalancer = True

        # LoadBalancerClassicExternal LoadBalancerClassicInternal
        cfg.LoadBalancerClassicExternal = None
        cfg.LoadBalancerClassicInternal = None
        try:
            cfg.LoadBalancerClassic
        except Exception:
            pass
        else:
            if 'External' in cfg.LoadBalancerClassic:
                cfg.LoadBalancerClassicExternal = True
            if 'Internal' in cfg.LoadBalancerClassic:
                cfg.LoadBalancerClassicInternal = True

        # LoadBalancerApplicationExternal LoadBalancerApplicationInternal
        cfg.LoadBalancerApplicationExternal = False
        cfg.LoadBalancerApplicationInternal = False
        try:
            cfg.LoadBalancerApplication
        except Exception:
            pass
        else:
            if 'External' in cfg.LoadBalancerApplication:
                cfg.LoadBalancerApplicationExternal = True
            if 'Internal' in cfg.LoadBalancerApplication:
                cfg.LoadBalancerApplicationInternal = True

        # RecordSet
        cfg.RecordSetExternal = None
        cfg.RecordSetInternal = None
        if 'External' in cfg.RecordSet:
            cfg.RecordSetExternal = True
        if 'Internal' in cfg.RecordSet:
            cfg.RecordSetInternal = True

# End inner methods

    cfg_role = [
        read_yaml(envrole, 'BASE', CFG_FILE_INT, stacktype),
        read_yaml(envrole, 'BASE', CFG_FILE_EXT, stacktype),
        read_yaml(envrole, brand, CFG_FILE_EXT, stacktype),
    ]

    cfgs = {
        'common': [
            read_yaml('common', 'BASE', CFG_FILE_INT),
            read_yaml('common', 'BASE', CFG_FILE_EXT),
            read_yaml('common', brand, CFG_FILE_EXT),
        ],
        'type': [
            read_yaml('TYPE', 'BASE', CFG_FILE_INT, stacktype),
            read_yaml('TYPE', 'BASE', CFG_FILE_EXT, stacktype),
            read_yaml('TYPE', brand, CFG_FILE_EXT, stacktype),
        ],
        'role': cfg_role,
    }

    RP = get_RP(cfgs)

    # print(RP['dev']['eu-west-1']['CloudFrontCacheBehaviors']
    #    [2]['QueryStringCacheKeys'])
    if cfg.debug:
        print('##########RP#########START#####')
        pprint(RP)
        print('##########RP#########END#######')

        print('##########EXCLUDED#######START#####')
        pprint(LD_EXCLUDED)
        print('##########EXCLUDED#######END#######')

        print('##########INCLUDED#######START#####')
        pprint(LD_INCLUDED)
        print('##########INCLUDED#######END#######')

    cfg.RP = RP
    cfg.RP_cmm = RP['cmm']['cmm']

    set_cfg()



import troposphere.elasticache as cch

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .route53 import R53_RecordSetCCH


class CCHCacheSubnetGroupPrivate(cch.SubnetGroup):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Description = Sub('${EnvShort}-Private')
        self.SubnetIds = Split(',', get_expvalue('SubnetsPrivate'))


class CCHCacheSubnetGroupPublic(cch.SubnetGroup):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Description = Sub('${EnvShort}-Public')
        self.SubnetIds = Split(',', get_expvalue('SubnetsPublic'))


def CCH_Cache(key):
    # Resources
    R_Cache = cch.CacheCluster('ElastiCacheCacheCluster')
    R_Group = cch.ReplicationGroup('ElastiCacheReplicationGroup')

    auto_get_props(R_Cache, f'{key}Base')

    auto_get_props(R_Group, f'{key}Base')
    auto_get_props(R_Group, 'ReplicationGroupBase')

    R53_RecordSetCCH()

    add_obj([
        R_Cache,
        R_Group])


def CCH_SubnetGroups(key):
    # Resources
    R_Private = CCHCacheSubnetGroupPrivate('CacheSubnetGroupPrivate')

    R_Public = CCHCacheSubnetGroupPublic('CacheSubnetGroupPublic')

    # Outputs
    O_Private = Output('CacheSubnetGroupPrivate')
    O_Private.Value = Ref('CacheSubnetGroupPrivate')
    O_Private.Export = Export('CacheSubnetGroupPrivate')

    O_Public = Output('CacheSubnetGroupPublic')
    O_Public.Value = Ref('CacheSubnetGroupPublic')
    O_Public.Export = Export('CacheSubnetGroupPublic')

    add_obj([
        R_Private,
        R_Public,
        O_Private,
        O_Public])



import troposphere.cloudfront as clf

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj, change_obj_data,
                     clf_compute_order)


class CFOriginAccessIdentity(clf.CloudFrontOriginAccessIdentity):
    def __init__(self, title, comment, **kwargs):
        super().__init__(title, **kwargs)
        self.CloudFrontOriginAccessIdentityConfig = (
            clf.CloudFrontOriginAccessIdentityConfig(Comment=comment))

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class CFDefaultCacheBehavior(clf.DefaultCacheBehavior):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title
        # Look for default values in awsibox/cfg.py [BASE_CFGS]
        auto_get_props(self)

        # Use CachePolicyId/OriginRequestPolicyId or legacy mode
        if 'CachePolicyId' in key:
            for k in ['DefaultTTL', 'MaxTTL', 'MinTTL', 'ForwardedValues']:
                try:
                    del self.properties[k]
                except Exception:
                    pass

        if 'TargetOriginId' in key:
            self.TargetOriginId = get_endvalue(f'{name}TargetOriginId')
        else:
            self.TargetOriginId = get_endvalue(
                f'CloudFrontCacheBehaviors0TargetOriginId')

        if 'LambdaFunctionARN' in key:
            condname = f'{name}LambdaFunctionARN'
            eventType = 'origin-request'
            if 'LambdaEventType' in key:
                eventType = key['LambdaEventType']
            # conditions
            add_obj(
                get_condition(condname, 'not_equals', 'None')
            )

            self.LambdaFunctionAssociations = [
                If(
                    condname,
                    clf.LambdaFunctionAssociation(
                        EventType=eventType,
                        LambdaFunctionARN=get_endvalue(condname)
                    ),
                    Ref('AWS::NoValue')
                )
            ]


class CFCacheBehavior(clf.CacheBehavior, CFDefaultCacheBehavior):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title
        self.PathPattern = get_endvalue(f'{name}PathPattern')


def CFCustomErrors():
    CustomErrorResponses = []
    mapname = 'CloudFrontCustomErrorResponses'

    try:
        ErrorResponses = getattr(cfg, mapname)
    except Exception:
        pass
    else:
        for n, v in ErrorResponses.items():
            resname = f'{mapname}{n}'
            CustomErrorResponse = clf.CustomErrorResponse(resname)
            CustomErrorResponses.append(
                auto_get_props(CustomErrorResponse))

    return CustomErrorResponses


def CFOrigins():
    Origins = []
    for n, v in cfg.CloudFrontOrigins.items():
        resname = f'CloudFrontOrigins{n}'

        try:
            S3OriginConfig = v['S3OriginConfig']
        except Exception:
            pass
        else:
            if 'OriginAccessIdentity' in S3OriginConfig:
                try:
                    del getattr(cfg, resname)['CustomOriginConfig']
                except Exception:
                    pass
            else:
                del getattr(cfg, resname)['S3OriginConfig']

        Origins.append(
            auto_get_props(clf.Origin(resname)))

    return Origins


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def CF_CloudFront(key):
    # Resources
    CloudFrontDistribution = clf.Distribution('CloudFrontDistribution')

    DistributionConfig = clf.DistributionConfig(
        'CloudFrontDistributionConfig')
    auto_get_props(DistributionConfig, 'CloudFrontDistributionIBOX_BASE')
    DistributionConfig.DefaultCacheBehavior = CFDefaultCacheBehavior(
        'CloudFrontCacheBehaviors0', key=cfg.CloudFrontCacheBehaviors['0'])

    cachebehaviors = []
    # Skip DefaultBehaviour
    itercachebehaviors = iter(cfg.CloudFrontCacheBehaviors.items())
    next(itercachebehaviors)

    # Automatically compute Behaviour Order based on PathPattern
    cfg.dbg_clf_compute_order = {}
    sortedcachebehaviors = sorted(
        itercachebehaviors,
        key=lambda x_y: clf_compute_order(x_y[1]['PathPattern']))

    if cfg.debug:
        print('##########CLF_COMPUTE_ORDER#########START#######')
        for n, v in iter(sorted(cfg.dbg_clf_compute_order.items(),
                                key=lambda x_y: x_y[1])):
            print(f'{n} {v}\n')
        print('##########CLF_COMPUTE_ORDER#########END#######')

    for n, v in sortedcachebehaviors:
        name = f'CloudFrontCacheBehaviors{n}'

        cachebehavior = CFCacheBehavior(name, key=v)

        # Create and Use Condition
        # only if PathPattern Value differ between envs
        if f'{name}PathPattern' not in cfg.fixedvalues:
            # conditions
            add_obj(
                get_condition(
                    name, 'not_equals', 'none', f'{name}PathPattern'))

            cachebehaviors.append(
                If(
                    name,
                    cachebehavior,
                    Ref("AWS::NoValue")))
        else:
            cachebehaviors.append(cachebehavior)

    DistributionConfig.CacheBehaviors = cachebehaviors

    cloudfrontaliasextra = []
    for n, v in cfg.CloudFrontAliasExtra.items():
        name = f'CloudFrontAliasExtra{n}'

        # skip parameter and condition for Cdn, Env, Zone
        if v.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
            cloudfrontaliasextra.append(get_endvalue(name))
            continue

        # parameters
        p_Alias = Parameter(name)
        p_Alias.Description = (
            'Alias extra - ''empty for default based on env/role')

        add_obj(p_Alias)

        cloudfrontaliasextra.append(get_endvalue(name, condition=True))

        # conditions
        add_obj(get_condition(name, 'not_equals', 'none'))

    DistributionConfig.Aliases = cloudfrontaliasextra
    DistributionConfig.CustomErrorResponses = CFCustomErrors()
    CloudFrontDistribution.DistributionConfig = DistributionConfig

    CloudFrontDistribution.DistributionConfig.Origins = CFOrigins()
    CloudFrontDistribution.DistributionConfig.Comment = get_endvalue(
        'CloudFrontComment')

    try:
        cfg.CloudFrontDistributionCondition
    except Exception:
        pass
    else:
        CloudFrontDistribution.Condition = (
            cfg.CloudFrontDistributionCondition)

    R_CloudFrontDistribution = CloudFrontDistribution

    add_obj([
        R_CloudFrontDistribution])



import troposphere.s3 as s3

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj, get_dictvalue)
from .iam import (IAMRoleBucketReplica, IAMPolicyBucketReplica,
                  IAMPolicyStatement)
from .cloudfront import CFOriginAccessIdentity
from .lambdas import LambdaPermissionS3


class S3Bucket(s3.Bucket):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. BucketStatic
        # need to specify key cause there is a problem regarind Bucket Names
        # key in common.yml and Bucket dict (RP_to_cfg)
        auto_get_props(self, key=key)
        self.Condition = name
        self.BucketName = Sub(bucket_name)
        self.CorsConfiguration = If(
            f'{name}Cors',
            s3.CorsConfiguration(
                CorsRules=[
                    s3.CorsRules(
                        AllowedHeaders=['Authorization'],
                        AllowedMethods=['GET'],
                        AllowedOrigins=['*'],
                        MaxAge=3000
                    )
                ]
            ),
            Ref('AWS::NoValue')
        )

        self.VersioningConfiguration = If(
            f'{name}Versioning',
            s3.VersioningConfiguration(
                Status=get_endvalue(f'{name}Versioning')
            ),
            Ref('AWS::NoValue')
        )


class S3BucketPolicy(s3.BucketPolicy):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        if 'Condition' in key:
            self.Condition = key['Condition']
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }


def S3BucketPolicyStatementBase(bucket):
    statements = []
    statements.append({
        'Action': [
            's3:GetBucketLocation'
        ],
        'Effect': 'Allow',
        'Resource': Sub('arn:aws:s3:::%s' % bucket_name),
        'Principal': {
            'AWS': Sub('arn:aws:iam::${AWS::AccountId}:root')
        },
        'Sid': 'Base'
    })

    return statements


def S3BucketPolicyStatementReplica(bucket, resource):
    statements = []
    if_statements = []
    condition = f'{bucket}PolicyStatementReplicaPrincipal'
    statements.append({
        'Action': [
            's3:ReplicateObject',
            's3:ReplicateDelete',
            's3:ObjectOwnerOverrideToBucketOwner',
            ],
        'Effect': 'Allow',
        'Resource': resource,
        'Principal': {
            'AWS': [
                get_subvalue(
                    'arn:aws:iam::${1M}:root',
                    f'{bucket}PolicyStatementReplicaPrincipal')
            ]
        },
        'Sid': 'AllowReplica'
    })

    for s in statements:
        if_statements.append(
            If(
                condition,
                s,
                Ref('AWS::NoValue')
            )
        )

    return if_statements


def S3BucketPolicyStatementAllowGetObject(bucket, principal, sid):
    statements = []
    statements.append(
        {
            'Action': [
                's3:GetObject'
            ],
            'Effect': 'Allow',
            'Resource': [
                Sub('arn:aws:s3:::%s/*' % bucket_name)
            ],
            'Principal': {
                'AWS': principal
            },
            'Sid': sid
        },
    )

    return statements


def S3BucketPolicyStatementSes(bucket):
    statements = []
    statements.append(
        {
            'Action': [
                's3:PutObject'
            ],
            'Effect': 'Allow',
            'Resource': [
                Sub('arn:aws:s3:::%s/*' % bucket_name)
            ],
            'Principal': {
                'Service': 'ses.amazonaws.com'
            },
            'Condition': {
                'StringEquals': {
                    'aws:Referer': Ref('AWS::AccountId')
                },
            },
            'Sid': 'AllowSES'
        },
    )

    return statements


def S3BucketPolicyStatementRead(bucket, principal):
    statements = []
    if_statements = []
    condition = f'{bucket}PolicyRead'
    statements.append({
        'Action': [
            's3:ListBucket',
            's3:GetBucketLocation',
            's3:ListBucketMultipartUploads',
            's3:ListBucketVersions'
        ],
        'Effect': 'Allow',
        'Resource': [
            Sub('arn:aws:s3:::%s' % bucket_name)
        ],
        'Principal': {
            'AWS': principal
        },
        'Sid': 'AllowListBucket'
    })
    statements.append({
        'Action': [
            's3:GetObject',
            's3:ListMultipartUploadParts'
        ],
        'Effect': 'Allow',
        'Resource': [
            Sub('arn:aws:s3:::%s/*' % bucket_name)
        ],
        'Principal': {
            'AWS': principal
        },
        'Sid': 'AllowGetObject'
    })

    for s in statements:
        if_statements.append(
            If(
                condition,
                s,
                Ref('AWS::NoValue')
            )
        )

    return if_statements


def S3BucketPolicyStatementWrite(bucket, principal):
    statements = []
    if_statements = []
    condition = f'{bucket}PolicyWrite'
    statements.append({
        'Action': [
            's3:Put*',
        ],
        'Effect': 'Allow',
        'Resource': [
            Sub('arn:aws:s3:::%s/*' % bucket_name)
        ],
        'Principal': {
            'AWS': principal
        },
        'Sid': 'AllowPut'
    })

    for s in statements:
        if_statements.append(
            If(
                condition,
                s,
                Ref('AWS::NoValue')
            )
        )

    return if_statements


def S3_Buckets(key):
    global bucket_name

    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        name = n
        bucket_name = getattr(cfg, resname)

        PolicyReadConditions = []
        PolicyReadPrincipal = []

        for m, w in v['AccountsRead'].items():
            accountread_name = f'{resname}AccountsRead{m}'
            # conditions
            add_obj(get_condition(accountread_name, 'not_equals', 'none'))

            PolicyReadConditions.append(Condition(accountread_name))
            PolicyReadPrincipal.append(If(
                accountread_name,
                get_subvalue('arn:aws:iam::${1M}:root', accountread_name),
                Ref('AWS::NoValue')
            ))

        # conditions
        if PolicyReadConditions:
            c_PolicyRead = {f'{resname}PolicyRead': Or(
                Equals('1', '0'),
                Equals('1', '0'),
                *PolicyReadConditions
            )}
        else:
            c_PolicyRead = {
                f'{resname}PolicyRead': Equals('True', 'False')}

        PolicyWriteConditions = []
        PolicyWritePrincipal = []

        for m, w in v['AccountsWrite'].items():
            accountwrite_name = f'{resname}AccountsWrite{m}'
            # conditions
            add_obj(get_condition(accountwrite_name, 'not_equals', 'none'))

            PolicyWriteConditions.append(Condition(accountwrite_name))
            PolicyWritePrincipal.append(If(
                accountwrite_name,
                get_subvalue('arn:aws:iam::${1M}:root', accountwrite_name),
                Ref('AWS::NoValue')
            ))

        # conditions
        if PolicyWriteConditions:
            c_PolicyWrite = {f'{resname}PolicyWrite': Or(
                Equals('1', '0'),
                Equals('1', '0'),
                *PolicyWriteConditions
            )}
        else:
            c_PolicyWrite = {
                f'{resname}PolicyWrite': Equals('True', 'False')}

        c_Create = get_condition(
            resname, 'equals', 'yes', f'{resname}Create')

        c_Versioning = get_condition(
            f'{resname}Versioning', 'equals', 'Enabled')

        c_Cors = get_condition(
            f'{resname}Cors', 'equals', 'yes')

        c_Replica = {f'{resname}Replica': And(
            Condition(resname),
            get_condition(
                '', 'equals', 'yes', f'{resname}ReplicationEnabled')
        )}
        c_ReplicaPolicyStatement = get_condition(
            f'{resname}PolicyStatementReplicaPrincipal',
            'not_equals', 'none')

        add_obj([
            c_PolicyRead,
            c_PolicyWrite,
            c_Create,
            c_Versioning,
            c_Cors,
            c_Replica,
            c_ReplicaPolicyStatement])

        # resources
        BucketPolicyStatement = []

        r_Bucket = S3Bucket(resname, key=v)

        Replica_Rules = []
        for m, w in v['Replication']['ConfigurationRules'].items():
            replica_name = f'{resname}ReplicationConfigurationRules{m}'

            # parameters
            p_replicabucket = Parameter(
                f'{replica_name}DestinationBucket')
            p_replicabucket.Description = ('Replica Destination Bucket '
                                           '- empty for default based on '
                                           'Env/Roles/Region')
            add_obj(p_replicabucket)

            # conditions
            add_obj([
                get_condition(f'{replica_name}DestinationBucket',
                              'not_equals', 'none')])

            # resources
            rule = s3.ReplicationConfigurationRules(replica_name)
            auto_get_props(rule)

            rule.Status = 'Enabled'
            Replica_Rules.append(If(
                f'{replica_name}DestinationBucket',
                rule,
                Ref('AWS::NoValue')
            ))

        if Replica_Rules:
            ReplicationConfiguration = s3.ReplicationConfiguration(
                '', Role=GetAtt(f'Role{resname}Replica', 'Arn'),
                Rules=Replica_Rules)
            r_Bucket.ReplicationConfiguration = If(
                f'{resname}Replica',
                ReplicationConfiguration,
                Ref('AWS::NoValue'))

        PolicyStatementReplicaResources = []
        for m, w in v['PolicyStatementReplica']['Resource'].items():
            polstatname = f'{resname}PolicyStatementReplicaResource{m}'
            # conditions
            add_obj(
                get_condition(
                    f'{polstatname}Prefix', 'not_equals', 'none'))

            PolicyStatementReplicaResources.append(If(
                f'{polstatname}Prefix',
                get_subvalue('arn:aws:s3:::%s/${1M}*' % bucket_name,
                             f'{polstatname}Prefix'),
                Ref('AWS::NoValue')
            ))

        r_Policy = S3BucketPolicy(f'BucketPolicy{name}', key=v)
        r_Policy.Condition = resname
        r_Policy.Bucket = Ref(resname)
        r_Policy.PolicyDocument['Statement'] = BucketPolicyStatement

        # At least one statement must be always present,
        # create a simple one with no conditions
        BucketPolicyStatement.extend(
            S3BucketPolicyStatementBase(resname))

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementReplica(
                resname, PolicyStatementReplicaResources))

        r_Role = IAMRoleBucketReplica(f'Role{resname}Replica')

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementRead(resname, PolicyReadPrincipal))

        BucketPolicyStatement.extend(
            S3BucketPolicyStatementWrite(resname, PolicyWritePrincipal))

        r_IAMPolicyReplica = IAMPolicyBucketReplica(
            f'IAMPolicyReplicaBucket{name}',
            bucket=resname,
            bucket_name=bucket_name,
            mapname=f'{resname}ReplicationConfigurationRules',
            key=v['Replication']['ConfigurationRules'])

        try:
            lbd_confs = v['NotificationConfiguration']['LambdaConfigurations']
        except Exception:
            pass
        else:
            for lbd_n, lbd_v in lbd_confs.items():
                lambda_arn = eval(lbd_v['Function'])
                if 'Fn::GettAtt' in lambda_arn.data:
                    permname = '%s%s' % (
                        lambda_arn.data['Fn::GettAtt'].replace(
                            'Lambda', 'LambdaPermission'),
                        resname)
                else:
                    permname = 'LambdaPermission'

                r_LambdaPermission = LambdaPermissionS3(
                    permname, key=lambda_arn, source=resname)

                add_obj(r_LambdaPermission)

        try:
            bucket_policies = getattr(cfg, 'BucketPolicy')
        except Exception:
            pass
        else:
            for policy_name, policy_value in bucket_policies.items():
                BucketPolicyStatement.append(get_dictvalue(policy_value))

        if 'WebsiteConfiguration' in v:
            r_Bucket.WebsiteConfiguration = s3.WebsiteConfiguration(
                f'{resname}WebsiteConfiguration')
            auto_get_props(r_Bucket.WebsiteConfiguration)

        if 'PolicyStatement' in v:
            FixedStatements = []
            for fsn, fsv in v['PolicyStatement'].items():
                FixedStatement = IAMPolicyStatement(fsv)
                FixedStatement['Principal'] = {
                    'AWS': eval(fsv['Principal'])
                }
                FixedStatement['Sid'] = fsv['Sid']
                FixedStatements.append(FixedStatement)
            BucketPolicyStatement.extend(FixedStatements)

        if 'PolicyStatementExGetObjectPrincipal' in v:
            BucketPolicyStatement.extend(
                S3BucketPolicyStatementAllowGetObject(
                    resname,
                    get_endvalue(
                        f'{resname}PolicyStatementExGetObjectPrincipal'),
                    'AllowGetObjectExPrincipal'))

        PolicyCloudFrontOriginAccessIdentityPrincipal = []
        if 'CloudFrontOriginAccessIdentity' in v:
            identityname = v['CloudFrontOriginAccessIdentity']
            identityresname = (
                f'CloudFrontOriginAccessIdentity{identityname}')

            PolicyCloudFrontOriginAccessIdentityPrincipal.append(
                Sub('arn:aws:iam::cloudfront:user/'
                    'CloudFront Origin Access Identity ${%s}'
                    % identityresname)
            )

            for ixn, ixv in (
                    v['CloudFrontOriginAccessIdentityExtra'].items()):
                ixname = (
                    f'{resname}CloudFrontOriginAccessIdentityExtra{ixn}')
                # conditions
                add_obj(get_condition(ixname, 'not_equals', 'none'))

                PolicyCloudFrontOriginAccessIdentityPrincipal.append(If(
                    ixname,
                    get_subvalue(
                        'arn:aws:iam::cloudfront:user/'
                        'CloudFront Origin Access Identity ${1M}', ixname),
                    Ref('AWS::NoValue')
                ))

            # conditions
            identitycondname = f'{resname}CloudFrontOriginAccessIdentity'
            c_identity = get_condition(
                identitycondname, 'not_equals', 'none')

            add_obj(c_identity)

            # resources
            BucketPolicyStatement.extend(
                S3BucketPolicyStatementAllowGetObject(
                    resname,
                    PolicyCloudFrontOriginAccessIdentityPrincipal,
                    'AllowCFAccess'))

            r_OriginAccessIdentity = CFOriginAccessIdentity(
                identityresname, comment=identityname)
            r_OriginAccessIdentity.Condition = identitycondname

            add_obj([
                r_OriginAccessIdentity,
            ])

            # outputs
            o_OriginAccessIdentity = Output(identityresname)
            o_OriginAccessIdentity.Value = Ref(identityresname)
            o_OriginAccessIdentity.Condition = identitycondname

            add_obj(o_OriginAccessIdentity)

        add_obj([
            r_Bucket,
            r_Policy,
            r_IAMPolicyReplica,
            r_Role])

        # outputs
        outvaluebase = Sub(bucket_name)
        if 'OutputValueRegion' in v:
            condname = f'{resname}OutputValueRegion'
            # conditions
            add_obj(get_condition(condname, 'not_equals', 'AWSRegion'))

            outvaluebase = If(
                condname,
                Sub('${Region}-%s'
                    % bucket_name.replace('${AWS::Region}-', '', 1),
                    **{'Region': get_endvalue(condname)}),
                outvaluebase
            )

        o_Bucket = Output(resname)
        o_Bucket.Value = If(
            resname,
            Ref(resname),
            outvaluebase
        )
        if resname == 'BucketAppRepository':
            o_Bucket.Export = Export(resname)

        add_obj([
            o_Bucket])



import troposphere.events as eve

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionEvent


class EVENetworkConfiguration(eve.NetworkConfiguration):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.AwsVpcConfiguration = eve.AwsVpcConfiguration(
            SecurityGroups=cfg.SecurityGroupsImport,
            Subnets=Split(',', get_expvalue('SubnetsPrivate'))
        )


class EVEEcsParameters(eve.EcsParameters):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.LaunchType = get_endvalue('LaunchType')
        self.NetworkConfiguration = If(
            'NetworkModeAwsVpc',
            EVENetworkConfiguration(''),
            Ref('AWS::NoValue'),
        )
        self.TaskDefinitionArn = Ref('TaskDefinition')


def EVE_EventRules(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        # parameters
        p_State = Parameter(f'{resname}State')
        p_State.Description = (
            'Events Rule State - empty for default based on env/role')
        p_State.AllowedValues = ['', 'DISABLED', 'ENABLED']

        if 'ScheduleExpression' in v:
            p_ScheduleExpression = Parameter(
                f'{resname}ScheduleExpression')
            p_ScheduleExpression.Description = (
                'Events Rule Schedule - empty for default based on env/role')

            add_obj(p_ScheduleExpression)

        add_obj([
            p_State])

        # resources
        Targets = []
        need_ecsEventsRole = None
        for m, w in v['Targets'].items():
            targetname = f'{resname}Targets{m}'
            Target = eve.Target(targetname)

            if m.startswith('Lambda'):
                permname = '%s%s' % (
                    m.replace('Lambda', 'LambdaPermission'), resname)
                r_LambdaPermission = LambdaPermissionEvent(
                    permname, key=w, source=resname)
                if 'Condition' in v:
                    r_LambdaPermission.Condition = v['Condition']

                add_obj(r_LambdaPermission)
            if m.startswith('ECSCluster'):
                props = {
                    'Arn': get_subvalue(
                        'arn:aws:ecs:${AWS::Region}:${AWS::AccountId}:'
                        'cluster/${1E}', 'Cluster', stack='ClusterStack'),
                    'EcsParameters': EVEEcsParameters(''),
                    'RoleArn': get_expvalue('RoleECSEvents'),
                    'Id': f'Target{m}',
                }
                need_ecsEventsRole = True

                # add common "fixed" props
                auto_get_props(Target, rootdict=props)

            # add props found in yaml cfg
            auto_get_props(Target)
            Targets.append(Target)

        r_Rule = eve.Rule(resname)
        auto_get_props(r_Rule)
        r_Rule.Name = Sub('${AWS::StackName}-${EnvRole}-' f'Rule{n}')
        r_Rule.Targets = Targets

        # outputs
        o_State = Output(f'{resname}State')
        o_State.Value = get_endvalue(f'{resname}State')

        if 'ScheduleExpression' in v:
            o_ScheduleExpression = Output(
                f'{resname}ScheduleExpression')
            o_ScheduleExpression.Value = get_endvalue(
                f'{resname}ScheduleExpression')

            add_obj(o_ScheduleExpression)

        add_obj([
            r_Rule,
            o_State])



import troposphere.sns as sns

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionSNS
from .sqs import SQSQueuePolicy


class SNSSubscription(sns.SubscriptionResource):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)
        auto_get_props(self)


class SNSTopic(sns.Topic):
    def setup(self, name):
        self.DisplayName = Sub('${AWS::StackName}.${EnvRole}-SNS%s' % name)

##


def SNS_Topics(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        r_Topic = SNSTopic(resname)
        r_Topic.setup(name=n)
        if 'Condition' in v:
            r_Topic.Condition = v['Condition']

        add_obj(r_Topic)

        # outputs
        if v.get('Export'):
            o_Topic = Output(resname)
            o_Topic.Value = Ref(resname)
            o_Topic.Export = Export(resname)
            if 'Condition' in v:
                o_Topic.Condition = v['Condition']

            add_obj(o_Topic)


def SNS_Subscriptions(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        r_Subscription = SNSSubscription(resname, key=v)

        add_obj(r_Subscription)

        if v['Protocol'] == 'lambda':
            lambdaname = resname.split('Lambda')[1]
            topicname = resname.split('Lambda')[0].replace(key, 'SNS')
            permname = f'LambdaPermission{lambdaname}{topicname}'

            r_LambdaPermission = LambdaPermissionSNS(permname, key=v)
            # Propagate condition, if present, to permission
            if hasattr(r_Subscription, 'Condition'):
                r_LambdaPermission.Condition = r_Subscription.Condition

            add_obj(r_LambdaPermission)

        if v['Protocol'] == 'sqs':
            queuename = resname.split('SQS')[1]
            topicname = resname.split('SQS')[0].replace(key, 'SNS')
            queuepolicyname = f'SQSQueuePolicy{queuename}{topicname}'

            r_QueuePolicy = SQSQueuePolicy(queuepolicyname, key=v)

            add_obj(r_QueuePolicy)



import troposphere.route53 as r53

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class R53RecordSetEFS(r53.RecordSetType):
    def __init__(self, title, efsname, **kwargs):
        super().__init__(title, **kwargs)
        condname = f'EFSFileSystem{efsname}'
        self.Condition = condname
        self.HostedZoneId = Ref('HostedZonePrivate')
        self.Name = Sub('efs-%s.%s' % (efsname, cfg.HostedZoneNamePrivate))
        self.ResourceRecords = [
            Sub('${%s}.efs.${AWS::Region}.amazonaws.com' % condname)
        ]
        self.Type = 'CNAME'
        self.TTL = '300'


class R53RecordSetNSServiceDiscovery(r53.RecordSetType):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.HostedZoneId = Ref('HostedZoneEnv')
        self.Name = Sub('find.%s' % cfg.HostedZoneNameEnv)
        self.ResourceRecords = GetAtt('PublicDnsNamespace', 'NameServers')
        self.Type = 'NS'
        self.TTL = '300'


def R53_RecordSetEC2LoadBalancer():
    # Resources
    # RecordSet External
    if cfg.RecordSetExternal:
        R_External = r53.RecordSetType('RecordSetExternal')
        auto_get_props(R_External, 'R53RecordSetEC2LoadBalancerExternal')

        # LoadBalancerClassic
        if cfg.LoadBalancerClassicExternal:
            R_External.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerClassicExternal.DNSName}')
        elif cfg.LoadBalancerClassicInternal:
            R_External.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerClassicInternal.DNSName}')

        # LoadBalancerApplication
        if cfg.LoadBalancerApplicationExternal:
            R_External.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerApplicationExternal.DNSName}')
        elif cfg.LoadBalancerApplicationInternal:
            R_External.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerApplicationInternal.DNSName}')

        # outputs
        O_External = Output('RecordSetExternal')
        O_External.Value = Ref('RecordSetExternal')

        add_obj([
            R_External,
            O_External])

    # RecordSet Internal
    if cfg.RecordSetInternal:
        R_Internal = r53.RecordSetType('RecordSetInternal')
        auto_get_props(R_Internal, 'R53RecordSetEC2LoadBalancerInternal')

        # LoadBalancerClassic
        if cfg.LoadBalancerClassicInternal:
            R_Internal.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerClassicInternal.DNSName}')
        elif cfg.LoadBalancerClassicExternal:
            R_Internal.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerClassicExternal.DNSName}')

        # LoadBalancerApplication
        if cfg.LoadBalancerApplicationInternal:
            R_Internal.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerApplicationInternal.DNSName}')
        elif cfg.LoadBalancerApplicationExternal:
            R_Internal.AliasTarget.DNSName = Sub(
                'dualstack.${LoadBalancerApplicationExternal.DNSName}')

        # outputs
        O_Internal = Output('RecordSetInternal')
        O_Internal.Value = Ref('RecordSetInternal')

        add_obj([
            R_Internal,
            O_Internal])


def R53_RecordSetECSLoadBalancer():
    # Resources
    if cfg.RecordSetExternal:
        R_External = r53.RecordSetType('RecordSetExternal')

        if cfg.LoadBalancerApplicationExternal:
            auto_get_props(
                R_External,
                'R53RecordSetECSLoadBalancerTargetExternalExternal')
        else:
            auto_get_props(
                R_External,
                'R53RecordSetECSLoadBalancerTargetInternalExternal')

        # outputs
        O_External = Output('RecordSetExternal')
        O_External.Value = Ref('RecordSetExternal')

        add_obj([
            R_External,
            O_External])

    if cfg.RecordSetInternal:
        R_Internal = r53.RecordSetType('RecordSetInternal')

        if cfg.LoadBalancerApplicationInternal:
            auto_get_props(
                R_Internal,
                'R53RecordSetECSLoadBalancerTargetInternalInternal')
        else:
            auto_get_props(
                R_Internal,
                'R53RecordSetECSLoadBalancerTargetExternalInternal')

        # outputs
        O_Internal = Output('RecordSetInternal')
        O_Internal.Value = Ref('RecordSetInternal')

        add_obj([
            R_Internal,
            O_Internal])


def R53_RecordSetRDS(rds_resname):
    for n in ['External', 'Internal']:
        if n not in cfg.RecordSet:
            continue
        # resources
        r_Record = r53.RecordSetType(rds_resname)
        auto_get_props(r_Record, f'R53RecordSetRDS{n}')
        r_Record.title = f'RecordSet{n}'

        # outputs
        o_Record = Output(f'RecordSet{n}')
        o_Record.Value = Ref(f'RecordSet{n}')

        add_obj([
            r_Record,
            o_Record])


def R53_RecordSetCCH():
    for n in ['External', 'Internal']:
        if n not in cfg.RecordSet:
            continue
        # resources
        r_Record = r53.RecordSetType(f'RecordSet{n}')
        auto_get_props(r_Record, f'R53RecordSetCCH{n}')

        r_RecordReadOnly = r53.RecordSetType(f'RecordSet{n}RO')
        auto_get_props(r_RecordReadOnly, f'R53RecordSetCCHReadOnly{n}')

        # outputs
        o_Record = Output(f'RecordSet{n}')
        o_Record.Value = Ref(f'RecordSet{n}')
        o_Record.Condition = 'CacheEnabled'

        o_RecordReadOnly = Output(f'RecordSet{n}RO')
        o_RecordReadOnly.Value = Ref(f'RecordSet{n}RO')
        o_RecordReadOnly.Condition = 'ReplicationGroup'

        add_obj([
            r_Record,
            r_RecordReadOnly,
            o_Record,
            o_RecordReadOnly])


def R53_HostedZones(key):
    for n, v in getattr(cfg, key).items():
        mapname = f'{key}{n}'
        resname = v['ResourceName']
        output_zonename = resname.replace('HostedZone', 'HostedZoneName')
        output_zoneidname = resname.replace('HostedZone', 'HostedZoneId')
        # parameters
        if n.startswith('Public'):
            p_HostedZone = Parameter(f'{mapname}Enabled')
            p_HostedZone.Description = (
                f'Create Public {resname} - can be created in only one '
                'Region - empty for default based on env/role')

            p_HostedZoneId = Parameter(f'{mapname}Id')
            p_HostedZoneId.Description = (
                f'Id of Public {resname} - required in all Regions where'
                ' HostedZonePublicEnv is not created - '
                'empty for default based on env/role')

            add_obj([
                p_HostedZone,
                p_HostedZoneId])

            # conditions
            c_Enabled = get_condition(
                resname, 'equals', 'yes', f'{mapname}Enabled')

            add_obj(c_Enabled)

        # resources
        r_HostedZone = r53.HostedZone(resname)
        auto_get_props(r_HostedZone, mapname)

        if n.startswith('Public'):
            r_HostedZone.Condition = resname

        # outputs
        o_HostedZoneName = Output(output_zonename)
        # o_HostedZoneName.Value = Sub(cfg.HostedZoneNamePrivate)
        o_HostedZoneName.Value = get_endvalue(f'{mapname}Name')

        o_HostedZoneId = Output(output_zoneidname)
        o_HostedZoneId.Value = If(
            resname,
            Ref(resname),
            get_endvalue(f'{mapname}Id')
        ) if n.startswith('Public') else Ref(resname)
        o_HostedZoneId.Export = Export(output_zoneidname)

        add_obj([
            r_HostedZone,
            o_HostedZoneName,
            o_HostedZoneId])


def R53_RecordSets(key):
    for n, v in getattr(cfg, key).items():
        if not v['IBOX_ENABLED']:
            continue
        resname = f'RecordSet{n}'

        r_Record = r53.RecordSetType(resname)
        add_obj(auto_get_props(r_Record, f'R53{resname}'))



import troposphere.efs as efs

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .route53 import R53RecordSetEFS
from .securitygroup import SecurityGroup, SecurityGroupIngress


class EFSFileSystem(efs.FileSystem):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title
        # by defualt do not encrypt and lower cost performaceMode
        self.Encrypted = False
        self.PerformanceMode = 'generalPurpose'
        auto_get_props(self)


class EFSMountTarget(efs.MountTarget):
    def __init__(self, title, index, sgname, efsresname, **kwargs):
        super().__init__(title, **kwargs)
        name = self.title
        self.Condition = name
        self.FileSystemId = Ref(efsresname)
        self.SecurityGroups = [
            Ref(sgname)
        ]
        self.SubnetId = Select(
            str(index), Split(',', get_expvalue('SubnetsPrivate')))


# #################################
# ### START STACK INFRA CLASSES ###
# #################################


def EFS_FileStorage(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. EFSFileSystemWordPress
        recordname = f'RecordSetEFS{n}'
        sgservername = f'SecurityGroupEFSServer{n}'
        sgclientname = f'SecurityGroupEFS{n}'
        sginame = f'SecurityGroupIngressEFS{n}'
        for i in range(cfg.AZones['MAX']):
            mountname = f'EFSMountTarget{n}{i}'
            # conditions
            add_obj(
                {mountname: And(
                    Condition(resname),
                    Equals(
                        FindInMap(
                            'AvabilityZones',
                            Ref('AWS::Region'),
                            f'Zone{i}'),
                        'True')
                )}
            )

            # resources
            r_Mount = EFSMountTarget(
                mountname, index=i,
                sgname=sgservername, efsresname=resname)

            add_obj(r_Mount)
        # conditions
        add_obj(get_condition(
            resname, 'equals', 'yes', f'{resname}Enabled'))

        # resources
        r_File = EFSFileSystem(resname, key=v)

        if v['R53']:
            r_Record = R53RecordSetEFS(recordname, efsname=n)

            add_obj(r_Record)

        r_SGServer = SecurityGroup(sgservername)
        r_SGServer.Condition = resname
        r_SGServer.GroupDescription = f'Rule to access EFS FileSystem {n}'

        r_SGClient = SecurityGroup(sgclientname)
        r_SGClient.Condition = resname
        r_SGClient.GroupDescription = f'Enable access to EFS FileSystem {n}'

        SGIExtra = {
            'Name': sginame,
            'Condition': resname,
            'FromPort': 2049,
            'GroupId': 'Sub(\'${%s.GroupId}\')' % sgservername,
            'SourceSecurityGroupId': 'Ref(\'%s\')' % sgclientname,
            'ToPort': 2049
        }

        r_SGI = SecurityGroupIngress(sginame)
        auto_get_props(r_SGI, rootdict=SGIExtra)

        # outputs
        o_File = Output(resname)
        o_File.Condition = resname
        o_File.Value = Ref(resname)
        o_File.Export = Export(resname)

        o_SGClient = Output(sgclientname)
        o_SGClient.Condition = resname
        o_SGClient.Value = GetAtt(sgclientname, 'GroupId')
        o_SGClient.Export = Export(sgclientname)

        add_obj([
            r_File,
            r_SGServer,
            r_SGClient,
            r_SGI,
            o_File,
            o_SGClient])



import troposphere.ecs as ecs

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .securitygroup import (SecurityGroupEcsService,
                            SecurityGroupRuleEcsService)


def ECS_ContainerDefinition():
    Containers = []
    for n, v in cfg.ContainerDefinitions.items():
        name = f'ContainerDefinitions{n}'  # Ex. ContainerDefinitions1

        EnvValue_Out_String = []
        EnvValue_Out_Map = {}
        for m, w in v['Environment'].items():
            if m.startswith('Env'):
                continue
            envname = f'{name}Environment{m}'
            envkeyname = w['Name']
            # parameters
            p_EnvValue = Parameter(f'{envname}Value')
            p_EnvValue.Description = (
                f'{envkeyname} - empty for default based on env/role')

            # If key NoParam is present skip adding Parameters
            # (usefull as they have a limited max number)
            if 'NoParam' not in w:
                add_obj(p_EnvValue)

            EnvValue_Out_String.append(
                '%s=${%s}' % (envkeyname, envkeyname))
            EnvValue_Out_Map.update({
                envkeyname: get_endvalue(f'{envname}Value')
            })

        # resources
        Container = ecs.ContainerDefinition(name)
        auto_get_props(Container, indexname=n)

        if len(cfg.ContainerDefinitions) == 1:
            Container.Cpu = If(
                'CpuTask',
                get_endvalue('Cpu'),
                get_endvalue(f'{name}Cpu')
            )
            Container.Memory = If(
                'LaunchTypeFarGate',
                get_endvalue('Memory'),
                get_endvalue(f'{name}Memory')
            )

        Containers.append(Container)

        # outputs
        o_EnvValueOut = Output(f'{name}Environment')
        o_EnvValueOut.Value = Sub(
            ','.join(EnvValue_Out_String), **EnvValue_Out_Map)

        add_obj(o_EnvValueOut)

    return Containers


def ECS_TaskDefinition(key):
    # Resources
    R_TaskDefinition = ecs.TaskDefinition('TaskDefinitionBase')
    auto_get_props(R_TaskDefinition)
    R_TaskDefinition.ContainerDefinitions = ECS_ContainerDefinition()

    add_obj([
        R_TaskDefinition])


def ECS_Service(key):
    # Resources
    R_SG = SecurityGroupEcsService('SecurityGroupEcsService')
    if cfg.LoadBalancerApplicationExternal:
        SGRule = SecurityGroupRuleEcsService(scheme='External')
        R_SG.SecurityGroupIngress.append(SGRule)

    if cfg.LoadBalancerApplicationInternal:
        SGRule = SecurityGroupRuleEcsService(scheme='Internal')
        R_SG.SecurityGroupIngress.append(SGRule)

    add_obj(R_SG)

    for n, v in getattr(cfg, key).items():
        if not v['IBOX_ENABLED']:
            continue
        mapname = f'{key}{n}'

        # delete not used LoadBalancers configuration, so that auto_get_props
        # do not find it
        for m in ['External', 'Internal']:
            if not getattr(cfg, f'LoadBalancerApplication{m}'):
                del v['LoadBalancers'][m]
        if not v['LoadBalancers']:
            # delete if empty to maintain compatibility with previous conf
            del v['LoadBalancers']

        r_Service = ecs.Service(mapname)
        auto_get_props(r_Service)

        if cfg.LoadBalancerApplication:
            r_Service.Role = If(
                'NetworkModeAwsVpc',
                Ref('AWS::NoValue'),
                get_expvalue('RoleECSService'))

        # When creating a service that specifies multiple target groups,
        # the Amazon ECS service-linked role must be created.
        # The role is created by omitting the Role property
        # in AWS CloudFormation
        if (cfg.LoadBalancerApplicationExternal and
                cfg.LoadBalancerApplicationInternal):
            r_Service.Role = Ref('AWS::NoValue')

        add_obj(r_Service)
